from django.db.models import IntegerField, Transform

from django_mysql.utils import collapse_spaces


class SetLength(Transform):
    lookup_name = "len"
    output_field = IntegerField()

    # No str.count equivalent in MySQL :(
    expr = collapse_spaces(
        """
        (
            CHAR_LENGTH(%s) -
            CHAR_LENGTH(REPLACE(%s, ',', '')) +
            IF(CHAR_LENGTH(%s), 1, 0)
        )
    """
    )

    def as_sql(self, compiler, connection):
        lhs, params = compiler.compile(self.lhs)
        return self.expr % (lhs, lhs, lhs), params



from io import StringIO
from textwrap import dedent
from unittest import SkipTest, mock

import django
import pytest
from django.core.management import CommandError, call_command
from django.db import connection
from django.db.utils import ConnectionHandler
from django.test import SimpleTestCase, TestCase, TransactionTestCase

from django_mysql.management.commands.fix_datetime_columns import parse_create_table
from django_mysql.utils import connection_is_mariadb

# Can't use @override_settings to swap out DATABASES, instead just mock.patch
# a new ConnectionHandler into the command module

command_connections = (
    "django_mysql.management.commands.fix_datetime_columns.connections"
)

sqlite = ConnectionHandler({"default": {"ENGINE": "django.db.backends.sqlite3"}})


def run_it(*args, **kwargs):
    run_args = ["fix_datetime_columns"]
    run_args.extend(args)

    out = StringIO()
    run_kwargs = {"stdout": out, "skip_checks": True}
    run_kwargs.update(kwargs)

    call_command(*run_args, **run_kwargs)

    return out.getvalue()


class Datetime6TestMixin:
    @classmethod
    def setUpClass(cls):
        if connection_is_mariadb(connection) or connection.mysql_version[:2] < (5, 6):
            raise SkipTest("Django only uses datetime(6) columns on MySQL 5.6+")
        super().setUpClass()


class FixDatetimeColumnsTests(Datetime6TestMixin, TestCase):

    if django.VERSION >= (2, 2):
        databases = ["default", "other"]
    else:
        multi_db = True

    def test_nothing_by_default(self):
        assert run_it() == ""

    def test_nothing_by_default_alternative_connection(self):
        assert run_it("other") == ""

    def test_invalid_database(self):
        with pytest.raises(CommandError) as excinfo:
            run_it("bla")

        assert "does not exist" in str(excinfo.value)

    @mock.patch(command_connections, sqlite)
    def test_invalid_not_mysql(self):
        with pytest.raises(CommandError) as excinfo:
            run_it()
        assert "not a MySQL database connection" in str(excinfo.value)


class SlowFixDatetimeColumnsTests(Datetime6TestMixin, TransactionTestCase):
    def test_with_one_column_to_migrate(self):
        with connection.cursor() as cursor:
            cursor.execute(
                """
                ALTER TABLE testapp_author
                    MODIFY COLUMN birthday datetime DEFAULT NULL
            """
            )
            try:
                out = run_it()
            finally:
                cursor.execute(
                    """
                    ALTER TABLE testapp_author
                        MODIFY COLUMN birthday datetime(6) DEFAULT NULL
                """
                )
        assert out == dedent(
            """\
            ALTER TABLE `testapp_author`
                MODIFY COLUMN `birthday` datetime(6) DEFAULT NULL;
            """
        )

    def test_with_two_columns_to_migrate(self):
        with connection.cursor() as cursor:
            cursor.execute(
                """
                ALTER TABLE testapp_author
                    MODIFY COLUMN birthday datetime DEFAULT NULL,
                    MODIFY COLUMN deathday datetime DEFAULT NULL
            """
            )
            try:
                out = run_it()
            finally:
                cursor.execute(
                    """
                    ALTER TABLE testapp_author
                        MODIFY COLUMN birthday datetime(6) DEFAULT NULL,
                        MODIFY COLUMN deathday datetime(6) DEFAULT NULL
                """
                )
        assert out == dedent(
            """\
            ALTER TABLE `testapp_author`
                MODIFY COLUMN `birthday` datetime(6) DEFAULT NULL,
                MODIFY COLUMN `deathday` datetime(6) DEFAULT NULL;
            """
        )


class ParseCreateTableTests(SimpleTestCase):
    def test_large_normalish_table(self):
        sql = (
            "CREATE TABLE `example` (\n"
            + "  `id` int(11) NOT NULL AUTO_INCREMENT,\n"
            + "  `varchary` varchar(255) NOT NULL,\n"
            + "  `texty` longtext NOT NULL,\n"
            + "  `inty` int(11) DEFAULT NULL,\n"
            + "  `datetimey` datetime NOT NULL,\n"
            + "  PRIMARY KEY (`id`),\n"
            + "  KEY `example_abcdef` (`varchary`),\n"
            + "  KEY `example_123456` (`inty`)\n"
            + "  CONSTRAINT `example_constraint` FOREIGN KEY (`id`) "
            + "REFERENCES `other` (`id`)\n"
            + ") ENGINE=InnoDB AUTO_INCREMENT=999999 DEFAULT CHARSET=utf8mb4 "
            + "ROW_FORMAT=COMPRESSED\n"
        )

        assert parse_create_table(sql) == {
            "id": "int(11) NOT NULL AUTO_INCREMENT",
            "varchary": "varchar(255) NOT NULL",
            "texty": "longtext NOT NULL",
            "inty": "int(11) DEFAULT NULL",
            "datetimey": "datetime NOT NULL",
        }

    def test_single_column(self):
        sql = (
            "CREATE TABLE `example_table` (\n"
            + "  `the_data` longtext COLLATE utf8mb4_unicode_ci NOT NULL\n"
            + ") ENGINE=MyISAM DEFAULT CHARSET=utf8mb4 "
            + "COLLATE=utf8mb4_unicode_ci COMMENT='hi mum';"
        )

        assert parse_create_table(sql) == {
            "the_data": "longtext COLLATE utf8mb4_unicode_ci NOT NULL"
        }



import pickle
import re
import zlib
from random import random
from textwrap import dedent
from time import time

from django.core.cache.backends.base import DEFAULT_TIMEOUT, BaseCache, default_key_func
from django.db import connections, router
from django.utils.encoding import force_bytes
from django.utils.module_loading import import_string

from django_mysql.utils import collapse_spaces, get_list_sql

BIGINT_SIGNED_MIN = -9223372036854775808
BIGINT_SIGNED_MAX = 9223372036854775807
BIGINT_UNSIGNED_MAX = 18446744073709551615


# Slightly modified copies of Options/BaseDatabaseCache from django's
# cache.backends.db - these allow us to act like a separate app for database
# routers (django_mysql), and not appear on django's `createcachetable`
# command


class Options:
    """A class that will quack like a Django model _meta class.

    This allows cache operations to be controlled by the router
    """

    def __init__(self, table):
        self.db_table = table
        self.app_label = "django_mysql"
        self.model_name = "cacheentry"
        self.verbose_name = "cache entry"
        self.verbose_name_plural = "cache entries"
        self.object_name = "CacheEntry"
        self.abstract = False
        self.managed = True
        self.proxy = False
        self.swapped = False


class BaseDatabaseCache(BaseCache):
    def __init__(self, table, params):
        super().__init__(params)
        self._table = table

        class CacheEntry:
            _meta = Options(table)

        self.cache_model_class = CacheEntry


reverse_key_re = re.compile(r"^([^:]*):(\d+):(.*)")


def default_reverse_key_func(full_key):
    """
    Reverse of Django's default_key_func, i.e. undoing:

        def default_key_func(key, key_prefix, version):
            return '%s:%s:%s' % (key_prefix, version, key)
    """
    match = reverse_key_re.match(full_key)
    return match.group(3), match.group(1), int(match.group(2))


def get_reverse_key_func(reverse_key_func):
    """
    Function to decide which reverse key function to use

    Defaults to ``None``, as any other value might not apply to the given
    KEY_FUNCTION. Also the user may not use any of the operations that require
    reversing the key_func.
    """
    if reverse_key_func is not None:
        if callable(reverse_key_func):
            return reverse_key_func
        else:
            return import_string(reverse_key_func)
    return None


class MySQLCache(BaseDatabaseCache):

    # Got an error with the add() query using BIGINT_UNSIGNED_MAX, so use a
    # value slightly 1 bit less (still an incalculable time into the future of
    # 1970)
    FOREVER_TIMEOUT = BIGINT_UNSIGNED_MAX >> 1

    create_table_sql = dedent(
        """\
        CREATE TABLE `{table_name}` (
            cache_key varchar(255) CHARACTER SET utf8 COLLATE utf8_bin
                                   NOT NULL PRIMARY KEY,
            value longblob NOT NULL,
            value_type char(1) CHARACTER SET latin1 COLLATE latin1_bin
                               NOT NULL DEFAULT 'p',
            expires BIGINT UNSIGNED NOT NULL
        );
    """
    )

    @classmethod
    def _now(cls):
        # Values in the expires column are milliseconds since unix epoch (UTC)
        return int(time() * 1000)

    def __init__(self, table, params):
        super().__init__(table, params)
        options = params.get("OPTIONS", {})
        self._compress_min_length = options.get("COMPRESS_MIN_LENGTH", 5000)
        self._compress_level = options.get("COMPRESS_LEVEL", 6)
        self._cull_probability = options.get("CULL_PROBABILITY", 0.01)

        # Figure out our *reverse* key function
        if self.key_func is default_key_func:
            self.reverse_key_func = default_reverse_key_func
            if ":" in self.key_prefix:
                raise ValueError(
                    "Cannot use the default KEY_FUNCTION and "
                    "REVERSE_KEY_FUNCTION if you have a colon in your "
                    "KEY_PREFIX."
                )
        else:
            reverse_key_func = params.get("REVERSE_KEY_FUNCTION", None)
            self.reverse_key_func = get_reverse_key_func(reverse_key_func)

    # Django API + helpers

    def get(self, key, default=None, version=None):
        key = self.make_key(key, version=version)
        self.validate_key(key)
        db = router.db_for_read(self.cache_model_class)
        table = connections[db].ops.quote_name(self._table)

        with connections[db].cursor() as cursor:
            cursor.execute(self._get_query.format(table=table), (key, self._now()))
            row = cursor.fetchone()

        if row is None:
            return default
        else:
            value, value_type = row
            return self.decode(value, value_type)

    _get_query = collapse_spaces(
        """
        SELECT value, value_type
        FROM {table}
        WHERE cache_key = %s AND
              expires >= %s
    """
    )

    def get_many(self, keys, version=None):
        made_key_to_key = {self.make_key(key, version=version): key for key in keys}
        made_keys = list(made_key_to_key.keys())
        for key in made_keys:
            self.validate_key(key)

        db = router.db_for_read(self.cache_model_class)
        table = connections[db].ops.quote_name(self._table)

        with connections[db].cursor() as cursor:
            cursor.execute(
                self._get_many_query.format(
                    table=table, list_sql=get_list_sql(made_keys)
                ),
                made_keys + [self._now()],
            )
            rows = cursor.fetchall()

        data = {}

        for made_key, value, value_type in rows:
            key = made_key_to_key[made_key]
            data[key] = self.decode(value, value_type)

        return data

    _get_many_query = collapse_spaces(
        """
        SELECT cache_key, value, value_type
        FROM {table}
        WHERE cache_key IN {list_sql} AND
              expires >= %s
    """
    )

    def set(self, key, value, timeout=DEFAULT_TIMEOUT, version=None):
        key = self.make_key(key, version=version)
        self.validate_key(key)
        self._base_set("set", key, value, timeout)

    def add(self, key, value, timeout=DEFAULT_TIMEOUT, version=None):
        key = self.make_key(key, version=version)
        self.validate_key(key)
        return self._base_set("add", key, value, timeout)

    def _base_set(self, mode, key, value, timeout=DEFAULT_TIMEOUT):
        if mode not in ("set", "add"):
            raise ValueError("'mode' should be 'set' or 'add'")

        exp = self.get_backend_timeout(timeout)
        db = router.db_for_write(self.cache_model_class)
        table = connections[db].ops.quote_name(self._table)

        self._maybe_cull()
        with connections[db].cursor() as cursor:

            value, value_type = self.encode(value)

            if mode == "set":
                query = self._set_query
                params = (key, value, value_type, exp)
            else:  # mode = 'add'
                query = self._add_query
                params = (key, value, value_type, exp, self._now())

            cursor.execute(query.format(table=table), params)

            if mode == "set":
                return True
            else:  # mode = 'add'
                # Use a special code in the add query for "did insert"
                insert_id = cursor.lastrowid
                return insert_id != 444

    _set_many_query = collapse_spaces(
        """
        INSERT INTO {table} (cache_key, value, value_type, expires)
        VALUES {{VALUES_CLAUSE}}
        ON DUPLICATE KEY UPDATE
            value=VALUES(value),
            value_type=VALUES(value_type),
            expires=VALUES(expires)
    """
    )

    _set_query = _set_many_query.replace("{{VALUES_CLAUSE}}", "(%s, %s, %s, %s)")

    # Uses the IFNULL / LEAST / LAST_INSERT_ID trick to communicate the special
    # value of 444 back to the client (LAST_INSERT_ID is otherwise 0, since
    # there is no AUTO_INCREMENT column)
    _add_query = collapse_spaces(
        """
        INSERT INTO {table} (cache_key, value, value_type, expires)
        VALUES (%s, %s, %s, %s)
        ON DUPLICATE KEY UPDATE
            value=IF(expires > @tmp_now:=%s, value, VALUES(value)),
            value_type=IF(expires > @tmp_now, value_type, VALUES(value_type)),
            expires=IF(
                expires > @tmp_now,
                IFNULL(
                    LEAST(LAST_INSERT_ID(444), NULL),
                    expires
                ),
                VALUES(expires)
            )
    """
    )

    def set_many(self, data, timeout=DEFAULT_TIMEOUT, version=None):
        exp = self.get_backend_timeout(timeout)
        db = router.db_for_write(self.cache_model_class)
        table = connections[db].ops.quote_name(self._table)

        self._maybe_cull()

        params = []
        for key, value in data.items():
            made_key = self.make_key(key, version=version)
            self.validate_key(made_key)
            value, value_type = self.encode(value)
            params.extend((made_key, value, value_type, exp))

        query = self._set_many_query.replace(
            "{{VALUES_CLAUSE}}", ",".join("(%s, %s, %s, %s)" for key in data)
        ).format(table=table)

        with connections[db].cursor() as cursor:
            cursor.execute(query, params)
        return []

    def delete(self, key, version=None):
        key = self.make_key(key, version=version)
        self.validate_key(key)

        db = router.db_for_write(self.cache_model_class)
        table = connections[db].ops.quote_name(self._table)

        with connections[db].cursor() as cursor:
            cursor.execute(self._delete_query.format(table=table), (key,))

    _delete_query = collapse_spaces(
        """
        DELETE FROM {table}
        WHERE cache_key = %s
    """
    )

    def delete_many(self, keys, version=None):
        made_keys = [self.make_key(key, version=version) for key in keys]
        for key in made_keys:
            self.validate_key(key)

        db = router.db_for_write(self.cache_model_class)
        table = connections[db].ops.quote_name(self._table)

        with connections[db].cursor() as cursor:
            cursor.execute(
                self._delete_many_query.format(
                    table=table, list_sql=get_list_sql(made_keys)
                ),
                made_keys,
            )

    _delete_many_query = collapse_spaces(
        """
        DELETE FROM {table}
        WHERE cache_key IN {list_sql}
    """
    )

    def has_key(self, key, version=None):
        key = self.make_key(key, version=version)
        self.validate_key(key)

        db = router.db_for_read(self.cache_model_class)
        table = connections[db].ops.quote_name(self._table)

        with connections[db].cursor() as cursor:
            cursor.execute(self._has_key_query.format(table=table), (key, self._now()))
            return cursor.fetchone() is not None

    _has_key_query = collapse_spaces(
        """
        SELECT 1 FROM {table}
        WHERE cache_key = %s and expires > %s
    """
    )

    def incr(self, key, delta=1, version=None):
        return self._base_delta(key, delta, version, "+")

    def decr(self, key, delta=1, version=None):
        return self._base_delta(key, delta, version, "-")

    def _base_delta(self, key, delta, version, operation):
        key = self.make_key(key, version=version)
        self.validate_key(key)

        db = router.db_for_write(self.cache_model_class)
        table = connections[db].ops.quote_name(self._table)

        with connections[db].cursor() as cursor:
            updated = cursor.execute(
                self._delta_query.format(table=table, operation=operation), (delta, key)
            )

            if not updated:
                raise ValueError("Key '%s' not found, or not an integer" % key)

            # New value stored in insert_id
            return cursor.lastrowid

    # Looks a bit tangled to turn the blob back into an int for updating, but
    # it works. Stores the new value for insert_id() with LAST_INSERT_ID
    _delta_query = collapse_spaces(
        """
        UPDATE {table}
        SET value = LAST_INSERT_ID(
            CAST(value AS SIGNED INTEGER)
            {operation}
            %s
        )
        WHERE cache_key = %s AND
              value_type = 'i'
    """
    )

    def clear(self):
        db = router.db_for_write(self.cache_model_class)
        table = connections[db].ops.quote_name(self._table)
        with connections[db].cursor() as cursor:
            cursor.execute("DELETE FROM {table}".format(table=table))

    def touch(self, key, timeout=DEFAULT_TIMEOUT, version=None):
        key = self.make_key(key, version=version)
        self.validate_key(key)
        exp = self.get_backend_timeout(timeout)
        db = router.db_for_write(self.cache_model_class)
        table = connections[db].ops.quote_name(self._table)
        with connections[db].cursor() as cursor:
            cursor.execute(
                self._touch_query.format(table=table), [exp, key, self._now()]
            )

    _touch_query = collapse_spaces(
        """
        UPDATE {table}
        SET expires = %s
        WHERE cache_key = %s AND
              expires >= %s
    """
    )

    def validate_key(self, key):
        """
        Django normally warns about maximum key length, but we error on it.
        """
        if len(key) > 250:
            raise ValueError(
                "Cache key is longer than the maxmimum 250 characters: {}".format(key)
            )
        return super().validate_key(key)

    def encode(self, obj):
        """
        Take a Python object and return it as a tuple (value, value_type), a
        blob and a one-char code for what type it is
        """
        if self._is_valid_mysql_bigint(obj):
            return obj, "i"

        value = pickle.dumps(obj, pickle.HIGHEST_PROTOCOL)
        value_type = "p"
        if self._compress_min_length and len(value) >= self._compress_min_length:
            value = zlib.compress(value, self._compress_level)
            value_type = "z"
        return value, value_type

    def _is_valid_mysql_bigint(self, value):
        return (
            # Can't support int subclasses since they should are expected to
            # decode back to the same object
            type(value) is int
            # Can't go beyond these ranges
            and BIGINT_SIGNED_MIN <= value <= BIGINT_SIGNED_MAX
        )

    def decode(self, value, value_type):
        """
        Take a value blob and its value_type one-char code and convert it back
        to a python object
        """
        if value_type == "i":
            return int(value)

        if value_type == "z":
            value = zlib.decompress(value)
            value_type = "p"

        if value_type == "p":
            return pickle.loads(force_bytes(value))

        raise ValueError(
            "Unknown value_type '{}' read from the cache table.".format(value_type)
        )

    def _maybe_cull(self):
        # Roll the dice, if it says yes then cull
        if self._cull_probability and random() <= self._cull_probability:
            self.cull()

    def get_backend_timeout(self, timeout=DEFAULT_TIMEOUT):
        if timeout is None:
            return self.FOREVER_TIMEOUT
        timeout = super().get_backend_timeout(timeout)
        return int(timeout * 1000)

    # Our API extensions

    def keys_with_prefix(self, prefix, version=None):
        if self.reverse_key_func is None:
            raise ValueError(
                "To use the _with_prefix commands with a custom KEY_FUNCTION, "
                "you need to specify a custom REVERSE_KEY_FUNCTION too."
            )

        if version is None:
            version = self.version

        db = router.db_for_read(self.cache_model_class)
        table = connections[db].ops.quote_name(self._table)

        prefix = self.make_key(prefix + "%", version=version)

        with connections[db].cursor() as cursor:
            cursor.execute(
                """SELECT cache_key FROM {table}
                   WHERE cache_key LIKE %s AND
                         expires >= %s""".format(
                    table=table
                ),
                (prefix, self._now()),
            )
            rows = cursor.fetchall()
            full_keys = {row[0] for row in rows}

            keys = {}
            for full_key in full_keys:
                key, key_prefix, key_version = self.reverse_key_func(full_key)

                if key_version == version:
                    keys[key] = key_version
            return set(keys.keys())

    def get_with_prefix(self, prefix, version=None):
        if self.reverse_key_func is None:
            raise ValueError(
                "To use the _with_prefix commands with a custom KEY_FUNCTION, "
                "you need to specify a custom REVERSE_KEY_FUNCTION too."
            )

        if version is None:
            version = self.version

        db = router.db_for_read(self.cache_model_class)
        table = connections[db].ops.quote_name(self._table)

        prefix = self.make_key(prefix + "%", version=version)
        version = str(version)

        with connections[db].cursor() as cursor:
            cursor.execute(
                """SELECT cache_key, value, value_type
                   FROM {table}
                   WHERE cache_key LIKE %s AND
                         expires >= %s""".format(
                    table=table
                ),
                (prefix, self._now()),
            )
            rows = cursor.fetchall()

            data = {}
            for made_key, value, value_type in rows:
                key, key_prefix, key_version = self.reverse_key_func(made_key)
                data[key] = self.decode(value, value_type)

            return data

    def delete_with_prefix(self, prefix, version=None):
        if version is None:
            version = self.version

        db = router.db_for_write(self.cache_model_class)
        table = connections[db].ops.quote_name(self._table)

        prefix = self.make_key(prefix + "%", version=version)

        with connections[db].cursor() as cursor:
            return cursor.execute(
                """DELETE FROM {table}
                   WHERE cache_key LIKE %s""".format(
                    table=table
                ),
                (prefix,),
            )

    def cull(self):
        db = router.db_for_write(self.cache_model_class)
        table = connections[db].ops.quote_name(self._table)

        with connections[db].cursor() as cursor:
            # First, try just deleting expired keys
            num_deleted = cursor.execute(
                "DELETE FROM {table} WHERE expires < %s".format(table=table),
                (self._now(),),
            )

            # -1 means "Don't limit size"
            if self._max_entries == -1:
                return

            cursor.execute("SELECT COUNT(*) FROM {table}".format(table=table))
            num = cursor.fetchone()[0]

            if num < self._max_entries:
                return num_deleted

            # Now do a key-based cull
            if self._cull_frequency == 0:
                num_deleted += cursor.execute("DELETE FROM {table}".format(table=table))
            else:
                cull_num = num // self._cull_frequency
                cursor.execute(
                    """SELECT cache_key FROM {table}
                       ORDER BY cache_key
                       LIMIT 1 OFFSET %s""".format(
                        table=table
                    ),
                    (cull_num,),
                )
                max_key = cursor.fetchone()[0]
                num_deleted += cursor.execute(
                    """DELETE FROM {table}
                       WHERE cache_key < %s""".format(
                        table=table
                    ),
                    (max_key,),
                )
            return num_deleted



import json
from datetime import date, datetime, time

import django
from django.db import connection
from django.db.models import (
    CASCADE,
    CharField,
    DateTimeField,
    DecimalField,
    ForeignKey,
    IntegerField,
)
from django.db.models import Model as VanillaModel
from django.db.models import OneToOneField, TextField
from django.utils import timezone

from django_mysql.models import (
    Bit1BooleanField,
    DynamicField,
    EnumField,
    JSONField,
    ListCharField,
    ListTextField,
    Model,
    NullBit1BooleanField,
    SetCharField,
    SetTextField,
    SizedBinaryField,
    SizedTextField,
)
from django_mysql.utils import connection_is_mariadb


class TemporaryModel(Model):
    """
    Used for temporary, mostly invalid models created in tests - check() is
    disabled unless an extra parameter is provided, in case the checks get run
    during tests, e.g. from call_command.
    """

    class Meta:
        app_label = "testapp"
        abstract = True

    @classmethod
    def check(cls, **kwargs):
        actually_check = kwargs.pop("actually_check", False)
        if actually_check:
            return super().check(**kwargs)
        else:
            return []


class EnumModel(Model):
    field = EnumField(
        choices=[
            ("red", "Red"),
            ("bloodred", "Blood Red"),
            "green",
            ("blue", "Navy Blue"),
            ("coralblue", "Coral Blue"),
        ]
    )


class NullableEnumModel(Model):
    field = EnumField(
        choices=["goat", ("deer", "Deer"), "bull", "dog's"],  # Test if escaping works
        null=True,
    )


class CharSetModel(Model):
    field = SetCharField(base_field=CharField(max_length=8), size=3, max_length=32)
    field2 = SetCharField(base_field=CharField(max_length=8), max_length=255)


class CharListModel(Model):
    field = ListCharField(base_field=CharField(max_length=8), size=3, max_length=32)


class IntSetModel(Model):
    field = SetCharField(base_field=IntegerField(), size=5, max_length=32)


class IntListModel(Model):
    field = ListCharField(base_field=IntegerField(), size=5, max_length=32)


class CharSetDefaultModel(Model):
    field = SetCharField(
        base_field=CharField(max_length=5),
        size=5,
        max_length=32,
        default=lambda: {"a", "d"},
    )


class CharListDefaultModel(Model):
    field = ListCharField(
        base_field=CharField(max_length=5),
        size=5,
        max_length=32,
        default=lambda: ["a", "d"],
    )


class BigCharSetModel(Model):
    field = SetTextField(base_field=CharField(max_length=8), max_length=32)


class BigCharListModel(Model):
    field = ListTextField(base_field=CharField(max_length=8))


class BigIntSetModel(Model):
    field = SetTextField(base_field=IntegerField())


class BigIntListModel(Model):
    field = ListTextField(base_field=IntegerField())


class DynamicModel(Model):
    attrs = DynamicField(
        spec={
            "datetimey": datetime,
            "datey": date,
            "floaty": float,
            "inty": int,
            "stry": str,
            "timey": time,
            "nesty": {"level2": str},
        }
    )

    @classmethod
    def check(cls, **kwargs):
        # Disable the checks on MySQL so that checks tests don't fail
        if not (
            connection_is_mariadb(connection) and connection.mysql_version >= (10, 0, 1)
        ):
            return []
        return super().check(**kwargs)

    def __unicode__(self):
        return ",".join("{}:{}".format(key, value) for key, value in self.attrs.items())


class SpeclessDynamicModel(Model):
    attrs = DynamicField()

    @classmethod
    def check(cls, **kwargs):
        # Disable the checks on MySQL so that checks tests don't fail
        if not (
            connection_is_mariadb(connection) and connection.mysql_version >= (10, 0, 1)
        ):
            return []
        return super().check(**kwargs)

    def __unicode__(self):
        return ",".join("{}:{}".format(key, value) for key, value in self.attrs.items())


class Author(Model):
    name = CharField(max_length=32, db_index=True)
    tutor = ForeignKey("self", on_delete=CASCADE, null=True, related_name="tutees")
    bio = TextField()
    birthday = DateTimeField(null=True)
    deathday = DateTimeField(null=True)

    def __unicode__(self):
        return "{} {}".format(self.id, self.name)


class Book(Model):
    title = CharField(max_length=32, db_index=True)
    author = ForeignKey(Author, on_delete=CASCADE, related_name="books")


class VanillaAuthor(VanillaModel):
    name = CharField(max_length=32)

    def __unicode__(self):
        return "{} {}".format(self.id, self.name)


class AuthorExtra(Model):
    author = OneToOneField(Author, on_delete=CASCADE, primary_key=True)
    legs = IntegerField(default=2)


class NameAuthor(Model):
    name = CharField(max_length=32, primary_key=True)

    def __unicode__(self):
        return "{} {}".format(self.id, self.name)


class NameAuthorExtra(Model):
    name = OneToOneField(NameAuthor, on_delete=CASCADE, primary_key=True)


class AuthorMultiIndex(Model):
    class Meta:
        index_together = ("name", "country")

    name = CharField(max_length=32, db_index=True)
    country = CharField(max_length=32)

    def __unicode__(self):
        return "{} {} in {}".format(self.id, self.name, self.country)


class AuthorHugeName(Model):
    class Meta:
        db_table = "this_is_an_author_with_an_incredibly_long_table_name_" "you_know_it"

    name = CharField(max_length=32)


class Alphabet(Model):
    a = IntegerField(default=1)
    b = IntegerField(default=2)
    c = IntegerField(default=3)
    d = CharField(max_length=32)
    e = CharField(max_length=32, null=True)
    f = IntegerField(null=True)
    g = DecimalField(default=0, decimal_places=2, max_digits=10)


class ProxyAlphabet(Alphabet):
    class Meta:
        proxy = True

    @property
    def a_times_b(self):
        return self.a * self.b


class Customer(Model):
    name = CharField(max_length=32)


class AgedCustomer(Customer):
    age = IntegerField(default=21)


class TitledAgedCustomer(AgedCustomer):
    class Meta:
        db_table = "titled aged customer"  # Test name quoting

    title = CharField(max_length=16)


class SizeFieldModel(Model):
    binary1 = SizedBinaryField(size_class=1)
    binary2 = SizedBinaryField(size_class=2)
    binary3 = SizedBinaryField(size_class=3)
    binary4 = SizedBinaryField(size_class=4)
    text1 = SizedTextField(size_class=1)
    text2 = SizedTextField(size_class=2)
    text3 = SizedTextField(size_class=3)
    text4 = SizedTextField(size_class=4)


class Bit1Model(Model):
    flag_a = Bit1BooleanField(default=True)
    flag_b = Bit1BooleanField(default=False)


class NullBit1Model(Model):
    flag = NullBit1BooleanField()


if django.VERSION >= (3, 1):
    with connection._nodb_cursor() as cursor:
        supports_json_field = (
            not cursor.db.mysql_is_mariadb and cursor.db.mysql_version >= (5, 7)
        )
else:
    nodb_conn = connection._nodb_connection
    try:
        supports_json_field = not connection_is_mariadb(
            nodb_conn
        ) and nodb_conn.mysql_version >= (5, 7)
    finally:
        nodb_conn.close()


class JSONModel(Model):
    # Something really weird and funky - change the fields on this model at
    # import time 😱
    if supports_json_field:
        attrs = JSONField(null=True)

    name = CharField(max_length=3)

    def __unicode__(self):
        return str(json.dumps(self.attrs))


# For cache tests


def expensive_calculation():
    expensive_calculation.num_runs += 1
    return timezone.now()


class Poll(VanillaModel):
    question = CharField(max_length=200)
    answer = CharField(max_length=200)
    pub_date = DateTimeField("date published", default=expensive_calculation)



from time import sleep
from unittest import SkipTest, mock, skipUnless

import django
import pytest
from django.db import DEFAULT_DB_ALIAS, connection, connections
from django.test import SimpleTestCase, TestCase

from django_mysql.utils import (
    PTFingerprintThread,
    WeightedAverageRate,
    connection_is_mariadb,
    format_duration,
    have_program,
    index_name,
    pt_fingerprint,
)
from tests.testapp.models import Author, AuthorMultiIndex

if django.VERSION < (3, 0):
    from django_mysql.utils import _is_mariadb_cache


class ConnectionIsMariaDBTests(TestCase):
    def setUp(self):
        if django.VERSION >= (3, 0):
            raise SkipTest("Not needed on Django 3.0+")
        super().setUp()
        _is_mariadb_cache.clear()

    def test_connection_proxy(self):
        connection_is_mariadb(connection)

    def test_connection(self):
        connection_is_mariadb(connections[DEFAULT_DB_ALIAS])

    def test_non_mysql(self):
        conn = mock.MagicMock(vendor="sqlite3")
        assert not connection_is_mariadb(conn)

    def test_oracle_mysql(self):
        conn = mock.MagicMock(vendor="mysql")
        conn.connection.get_server_info.return_value = "5.7.19"
        assert not connection_is_mariadb(conn)
        # check cached
        conn.connection.get_server_info.side_effect = ValueError("re-called")
        assert not connection_is_mariadb(conn)

    def test_mariadb(self):
        conn = mock.MagicMock(vendor="mysql")
        conn.connection.get_server_info.return_value = "10.0.3-MariaDB-1~precise-log"
        assert connection_is_mariadb(conn)
        conn.connection.get_server_info.side_effect = ValueError("re-called")
        assert connection_is_mariadb(conn)


class WeightedAverageRateTests(SimpleTestCase):
    def test_constant(self):
        # If we keep achieving a rate of 100 rows in 0.5 seconds, it should
        # recommend that we keep there
        rate = WeightedAverageRate(0.5)
        assert rate.update(100, 0.5) == 100
        assert rate.update(100, 0.5) == 100
        assert rate.update(100, 0.5) == 100

    def test_slow(self):
        # If we keep achieving a rate of 100 rows in 1 seconds, it should
        # recommend that we move to 50
        rate = WeightedAverageRate(0.5)
        assert rate.update(100, 1.0) == 50
        assert rate.update(100, 1.0) == 50
        assert rate.update(100, 1.0) == 50

    def test_fast(self):
        # If we keep achieving a rate of 100 rows in 0.25 seconds, it should
        # recommend that we move to 200
        rate = WeightedAverageRate(0.5)
        assert rate.update(100, 0.25) == 200
        assert rate.update(100, 0.25) == 200
        assert rate.update(100, 0.25) == 200

    def test_good_guess(self):
        # If we are first slow then hit the target at 50, we should be good
        rate = WeightedAverageRate(0.5)
        assert rate.update(100, 1.0) == 50
        assert rate.update(50, 0.5) == 50
        assert rate.update(50, 0.5) == 50
        assert rate.update(50, 0.5) == 50

    def test_zero_division(self):
        rate = WeightedAverageRate(0.5)
        assert rate.update(1, 0.0) == 500


class FormatDurationTests(SimpleTestCase):
    def test_seconds(self):
        assert format_duration(0) == "0s"
        assert format_duration(1) == "1s"
        assert format_duration(30) == "30s"
        assert format_duration(59) == "59s"

    def test_minutes(self):
        assert format_duration(60) == "1m0s"
        assert format_duration(61) == "1m1s"
        assert format_duration(120) == "2m0s"
        assert format_duration(3599) == "59m59s"

    def test_hours(self):
        assert format_duration(3600) == "1h0m0s"
        assert format_duration(3601) == "1h0m1s"


@skipUnless(have_program("pt-fingerprint"), "pt-fingerprint must be installed")
class PTFingerprintTests(SimpleTestCase):
    def test_basic(self):
        assert pt_fingerprint("SELECT 5") == "select ?"
        assert pt_fingerprint("SELECT 5;") == "select ?"

    def test_long(self):
        query = """
            SELECT
                CONCAT(customer.last_name, ', ', customer.first_name)
                    AS customer,
                address.phone,
                film.title
            FROM rental
                INNER JOIN customer
                    ON rental.customer_id = customer.customer_id
                INNER JOIN address
                    ON customer.address_id = address.address_id
                INNER JOIN inventory
                    ON rental.inventory_id = inventory.inventory_id
                INNER JOIN film
                    ON inventory.film_id = film.film_id
            WHERE
                rental.return_date IS NULL AND
                rental_date + INTERVAL film.rental_duration DAY <
                    CURRENT_DATE()
            LIMIT 5"""
        assert pt_fingerprint(query) == (
            "select concat(customer.last_name, ?, customer.first_name) as "
            + "customer, address.phone, film.title from rental inner join "
            + "customer on rental.customer_id = customer.customer_id "
            + "inner join address on customer.address_id = "
            + "address.address_id inner join inventory on "
            + "rental.inventory_id = inventory.inventory_id inner join "
            + "film on inventory.film_id = film.film_id where "
            + "rental.return_date is ? and rental_date ? interval "
            + "film.rental_duration day < current_date() limit ?"
        )

    def test_the_thread_shuts_on_time_out(self):
        PTFingerprintThread.PROCESS_LIFETIME = 0.1
        pt_fingerprint("select 123")
        sleep(0.2)
        assert PTFingerprintThread.the_thread is None
        PTFingerprintThread.PROCESS_LIFETIME = 60


class IndexNameTests(TestCase):

    if django.VERSION >= (2, 2):
        databases = ["default", "other"]
    else:
        multi_db = True

    def test_requires_field_names(self):
        with pytest.raises(ValueError) as excinfo:
            index_name(Author)
        assert "At least one field name required" in str(excinfo.value)

    def test_requires_real_field_names(self):
        with pytest.raises(ValueError) as excinfo:
            index_name(Author, "nonexistent")
        assert "Fields do not exist: nonexistent" in str(excinfo.value)

    def test_invalid_kwarg(self):
        with pytest.raises(ValueError) as excinfo:
            index_name(Author, "name", nonexistent_kwarg=True)
        assert "The only supported keyword argument is 'using'" in str(excinfo.value)

    def test_primary_key(self):
        assert index_name(Author, "id") == "PRIMARY"

    def test_primary_key_using_other(self):
        assert index_name(Author, "id", using="other") == "PRIMARY"

    def test_secondary_single_field(self):
        name = index_name(Author, "name")
        assert name.startswith("testapp_author_")

    def test_index_does_not_exist(self):
        with pytest.raises(KeyError) as excinfo:
            index_name(Author, "bio")
        assert "There is no index on (bio)" in str(excinfo.value)

    def test_secondary_multiple_fields(self):
        name = index_name(AuthorMultiIndex, "name", "country")
        assert name.startswith("testapp_authormultiindex")

    def test_secondary_multiple_fields_non_existent_reversed_existent(self):
        # Checks that order is preserved
        with pytest.raises(KeyError):
            index_name(AuthorMultiIndex, "country", "name")

    def test_secondary_multiple_fields_non_existent(self):
        with pytest.raises(KeyError):
            index_name(AuthorMultiIndex, "country", "id")



from django.db.migrations.operations.base import Operation
from django.utils.functional import cached_property


class InstallPlugin(Operation):
    reduces_to_sql = False

    reversible = True

    def __init__(self, name, soname):
        self.name = name
        self.soname = soname

    def state_forwards(self, app_label, state):
        pass  # pragma: no cover

    def database_forwards(self, app_label, schema_editor, from_st, to_st):
        if not self.plugin_installed(schema_editor):
            schema_editor.execute(
                "INSTALL PLUGIN {} SONAME %s".format(self.name), (self.soname,)
            )

    def database_backwards(self, app_label, schema_editor, from_st, to_st):
        if self.plugin_installed(schema_editor):
            schema_editor.execute("UNINSTALL PLUGIN %s" % self.name)

    def plugin_installed(self, schema_editor):
        with schema_editor.connection.cursor() as cursor:
            cursor.execute(
                """SELECT COUNT(*)
                FROM INFORMATION_SCHEMA.PLUGINS
                WHERE PLUGIN_NAME LIKE %s""",
                (self.name,),
            )
            count = cursor.fetchone()[0]
            return count > 0

    def describe(self):
        return "Installs plugin {} from {}".format(self.name, self.soname)


class InstallSOName(Operation):
    reduces_to_sql = True

    reversible = True

    def __init__(self, soname):
        self.soname = soname

    def state_forwards(self, app_label, state):
        pass  # pragma: no cover

    def database_forwards(self, app_label, schema_editor, from_st, to_st):
        schema_editor.execute("INSTALL SONAME %s", (self.soname,))

    def database_backwards(self, app_label, schema_editor, from_st, to_st):
        schema_editor.execute("UNINSTALL SONAME %s", (self.soname,))

    def describe(self):
        return "Installs library %s" % (self.soname)


class AlterStorageEngine(Operation):
    def __init__(self, name, to_engine, from_engine=None):
        self.name = name
        self.engine = to_engine
        self.from_engine = from_engine

    @property
    def reversible(self):
        return self.from_engine is not None

    def state_forwards(self, app_label, state):
        pass

    def database_forwards(
        self, app_label, schema_editor, from_state, to_state, engine=None
    ):
        if engine is None:
            engine = self.engine

        new_model = to_state.apps.get_model(app_label, self.name)

        qn = schema_editor.connection.ops.quote_name

        if self.allow_migrate_model(schema_editor.connection.alias, new_model):
            with schema_editor.connection.cursor() as cursor:
                cursor.execute(
                    """SELECT COUNT(*) FROM INFORMATION_SCHEMA.TABLES
                       WHERE TABLE_SCHEMA=DATABASE() AND
                             TABLE_NAME = %s AND
                             ENGINE = %s""",
                    (new_model._meta.db_table, engine),
                )
                uses_engine_already = cursor.fetchone()[0] > 0

            if uses_engine_already:
                return

            schema_editor.execute(
                "ALTER TABLE {table} ENGINE={engine}".format(
                    table=qn(new_model._meta.db_table), engine=engine
                )
            )

    def database_backwards(self, app_label, schema_editor, from_state, to_state):
        if self.from_engine is None:
            raise NotImplementedError("You cannot reverse this operation")

        return self.database_forwards(
            app_label, schema_editor, from_state, to_state, engine=self.from_engine
        )

    @cached_property
    def name_lower(self):
        return self.name.lower()

    def references_model(self, name, app_label=None):
        return name.lower() == self.name_lower

    def describe(self):
        if self.from_engine:
            from_clause = " from {}".format(self.from_engine)
        else:
            from_clause = ""
        return "Alter storage engine for {model}{from_clause} to {engine}".format(
            model=self.name, from_clause=from_clause, engine=self.engine
        )



import warnings
from contextlib import closing

import django
from django.db import connection
from pytest_django.plugin import _blocking_manager


def pytest_report_header(config):
    dot_version = ".".join(str(x) for x in django.VERSION)
    header = "Django version: " + dot_version

    with _blocking_manager.unblock():
        if django.VERSION >= (3, 1):

            with connection._nodb_cursor() as cursor:
                cursor.execute("SELECT VERSION()")
                version = cursor.fetchone()[0]

        else:

            with closing(connection._nodb_connection) as conn:
                with conn.cursor() as cursor:
                    cursor.execute("SELECT VERSION()")
                    version = cursor.fetchone()[0]
        header += "\nMySQL version: {}".format(version)

    return header


# MySQL 5.7 warns about some sql mode changes
warnings.filterwarnings("ignore", r".*Changing sql mode.*")
warnings.filterwarnings("ignore", r".*sql modes should be used with strict mode.*")
# MySQL 5.7 turned 'explain' into 'explain extended' so it always warns the
# optimized query
warnings.filterwarnings("ignore", r".*/\* select#\d+ \*/")
# MySQL 5.7 deprecated some query hints
warnings.filterwarnings("ignore", r".*'SQL_CACHE' is deprecated.*")
warnings.filterwarnings("ignore", r".*'SQL_NO_CACHE' is deprecated.*")



from django.db.models import CharField
from django.utils.encoding import force_str
from django.utils.translation import ugettext_lazy as _


class EnumField(CharField):
    description = _("Enumeration")

    def __init__(self, *args, **kwargs):
        if "choices" not in kwargs or len(kwargs["choices"]) == 0:
            raise ValueError('"choices" argument must be be a non-empty list')

        choices = []
        for choice in kwargs["choices"]:
            if isinstance(choice, tuple):
                choices.append(choice)
            elif isinstance(choice, str):
                choices.append((choice, choice))
            else:
                raise TypeError(
                    'Invalid choice "{choice}". '
                    "Expected string or tuple as elements in choices".format(
                        choice=choice
                    )
                )

        kwargs["choices"] = choices

        if "max_length" in kwargs:
            raise TypeError('"max_length" is not a valid argument')
        # Massive to avoid problems with validation - let MySQL handle the
        # maximum string length
        kwargs["max_length"] = int(2 ** 32)

        super().__init__(*args, **kwargs)

    def deconstruct(self):
        name, path, args, kwargs = super().deconstruct()

        bad_paths = (
            "django_mysql.models.fields.enum.EnumField",
            "django_mysql.models.fields.EnumField",
        )
        if path in bad_paths:
            path = "django_mysql.models.EnumField"

        kwargs["choices"] = self.choices
        del kwargs["max_length"]

        return name, path, args, kwargs

    def db_type(self, connection):
        connection.ensure_connection()
        values = [connection.connection.escape_string(c) for c, _ in self.flatchoices]
        # Use force_str because MySQLdb escape_string() returns bytes, but
        # pymysql returns str
        return "enum(%s)" % ",".join("'%s'" % force_str(v) for v in values)



import django
import pytest
from django.core.exceptions import ValidationError
from django.core.management import call_command
from django.db import connection
from django.db.utils import DataError
from django.test import TestCase, TransactionTestCase, override_settings

from django_mysql.models import EnumField
from tests.testapp.models import EnumModel, NullableEnumModel


class TestEnumField(TestCase):
    def test_empty_choices(self):
        with pytest.raises(ValueError):
            EnumField(choices=[])

    def test_no_choices(self):
        with pytest.raises(ValueError):
            EnumField()

    def test_invalid_choices(self):
        with pytest.raises(TypeError) as exc_info:
            EnumField(choices=["red", 10])
        assert 'Invalid choice "10"' in str(exc_info.value)

    def test_invalid_max_length(self):
        with pytest.raises(TypeError) as exc_info:
            EnumField(choices=["red"], max_length=100)
        assert '"max_length" is not a valid argument' in str(exc_info.value)

    def test_correct(self):
        s = EnumModel.objects.create(field="red")
        assert s.field == "red"
        s = EnumModel.objects.get(id=s.id)
        assert s.field == "red"

    def test_invalid_value(self):
        with pytest.raises(DataError):
            EnumModel.objects.create(field="elephant")

    def test_empty(self):
        with pytest.raises(DataError):
            EnumModel.objects.create()

    def test_null_storage(self):
        s = NullableEnumModel.objects.create()
        assert s.field is None
        s = NullableEnumModel.objects.get(id=s.id)
        assert s.field is None

    def test_isnull_lookup(self):
        NullableEnumModel.objects.create(field="goat")
        s = NullableEnumModel.objects.create()
        actual = NullableEnumModel.objects.filter(field__isnull=True)
        expected = [s]

        assert list(actual) == expected

    def test_basic_lookup(self):
        s1 = EnumModel.objects.create(field="green")
        EnumModel.objects.create(field="red")
        s2 = EnumModel.objects.create(field="green")

        actual = EnumModel.objects.filter(field="green")
        expected = [s1, s2]

        assert list(actual) == expected

    def test_in_lookup(self):
        s1 = EnumModel.objects.create(field="green")
        EnumModel.objects.create(field="red")
        s2 = EnumModel.objects.create(field="green")

        actual = EnumModel.objects.filter(field__in=["green"])
        expected = [s1, s2]

        assert list(actual) == expected

    def test_icontains_lookup(self):
        s1 = EnumModel.objects.create(field="coralBlue")
        s2 = EnumModel.objects.create(field="blue")
        EnumModel.objects.create(field="green")

        actual = EnumModel.objects.filter(field__icontains="Blue")
        expected = [s1, s2]

        assert list(actual) == expected

    def test_contains_lookup(self):
        s1 = EnumModel.objects.create(field="coralblue")
        s2 = EnumModel.objects.create(field="blue")
        EnumModel.objects.create(field="green")

        actual = EnumModel.objects.filter(field__contains="blue")
        expected = [s1, s2]

        assert list(actual) == expected


class TestCheck(TestCase):

    if django.VERSION >= (2, 2):
        databases = ["default", "other"]
    else:
        multi_db = True

    def test_check(self):
        errors = EnumModel.check()
        assert errors == []


class TestDeconstruct(TestCase):
    def test_deconstruct(self):
        field = EnumField(choices=["a", "b"])
        name, path, args, kwargs = field.deconstruct()
        assert path == "django_mysql.models.EnumField"
        assert "max_length" not in kwargs
        EnumField(*args, **kwargs)


class TestMigrations(TransactionTestCase):
    @override_settings(
        MIGRATION_MODULES={"testapp": "tests.testapp.enum_default_migrations"}
    )
    def test_adding_field_with_default(self):
        table_name = "testapp_enumdefaultmodel"
        table_names = connection.introspection.table_names
        with connection.cursor() as cursor:
            assert table_name not in table_names(cursor)

        call_command(
            "migrate", "testapp", verbosity=0, skip_checks=True, interactive=False
        )
        with connection.cursor() as cursor:
            assert table_name in table_names(cursor)

        call_command(
            "migrate",
            "testapp",
            "zero",
            verbosity=0,
            skip_checks=True,
            interactive=False,
        )
        with connection.cursor() as cursor:
            assert table_name not in table_names(cursor)


class TestFormfield(TestCase):
    def test_formfield(self):
        model_field = EnumField(choices=["this", "that"])
        form_field = model_field.formfield()

        assert form_field.clean("this") == "this"
        assert form_field.clean("that") == "that"

        with pytest.raises(ValidationError):
            form_field.clean("")

        with pytest.raises(ValidationError):
            form_field.clean("invalid")



from contextlib import contextmanager
from unittest import mock

import pytest
from django.db import DEFAULT_DB_ALIAS, connection, connections
from django.test.utils import CaptureQueriesContext

from django_mysql.utils import connection_is_mariadb


@contextmanager
def skip_if_mysql_8_plus():
    if not connection_is_mariadb(connection) and connection.mysql_version >= (8,):
        pytest.skip("Requires MySQL<8 or MariaDB")
    yield


def column_type(table_name, column_name):
    with connection.cursor() as cursor:
        cursor.execute(
            """SELECT DATA_TYPE FROM INFORMATION_SCHEMA.COLUMNS
               WHERE TABLE_SCHEMA = DATABASE() AND TABLE_NAME = %s AND
                     COLUMN_NAME = %s""",
            (table_name, column_name),
        )
        return cursor.fetchone()[0]


class CaptureLastQuery:
    def __init__(self, conn=None):
        if conn is None:
            self.conn = connection
        else:
            self.conn = conn

    def __enter__(self):
        self.capturer = CaptureQueriesContext(self.conn)
        self.capturer.__enter__()
        return self

    def __exit__(self, a, b, c):
        self.capturer.__exit__(a, b, c)

    @property
    def query(self):
        return self.capturer.captured_queries[-1]["sql"]


class print_all_queries:
    def __init__(self, conn=None):
        if conn is None:
            self.conn = connection
        else:
            self.conn = conn

    def __enter__(self):
        self.capturer = CaptureQueriesContext(self.conn)
        self.capturer.__enter__()
        return self

    def __exit__(self, a, b, c):
        self.capturer.__exit__(a, b, c)
        for q in self.capturer.captured_queries:
            print(q["sql"])


def used_indexes(query, using=DEFAULT_DB_ALIAS):
    """
    Given SQL 'query', run EXPLAIN and return the names of the indexes used
    """
    connection = connections[using]
    with connection.cursor() as cursor:
        cursor.execute("EXPLAIN " + query)

        return {row["key"] for row in fetchall_dicts(cursor) if row["key"] is not None}


def fetchall_dicts(cursor):
    columns = [x[0] for x in cursor.description]
    rows = []
    for row in cursor.fetchall():
        rows.append(dict(zip(columns, row)))
    return rows


@contextmanager
def mock_mysql_version(*, default, other):
    mock_default = mock.patch.object(
        connections["default"], "mysql_version", new=default
    )
    mock_other = mock.patch.object(connections["other"], "mysql_version", new=other)
    with mock_default, mock_other:
        yield



import sys

if sys.version_info >= (3, 7):
    from contextlib import nullcontext
else:

    class nullcontext:
        def __enter__(self):
            pass

        def __exit__(self, *exc_info):
            pass



from django.db import migrations, models

from django_mysql.models import ListCharField


class Migration(migrations.Migration):

    dependencies = [("testapp", "0001_initial")]

    operations = [
        migrations.AddField(
            model_name="intlistdefaultmodel",
            name="field_2",
            field=ListCharField(
                models.IntegerField(), default=lambda: [], size=None, max_length=32
            ),
        ),
        migrations.AddField(
            model_name="intlistdefaultmodel",
            name="field_3",
            field=ListCharField(
                models.IntegerField(), default=lambda: [1, 5], size=None, max_length=32
            ),
        ),
    ]



import json

from django.core import serializers
from django.db.models import F
from django.test import SimpleTestCase, TestCase

from tests.testapp.models import Bit1Model, NullBit1Model


class TestSaveLoad(TestCase):
    def test_basic(self):
        m = Bit1Model()
        m.flag_a = False
        m.flag_b = True
        m.save()

        m = Bit1Model.objects.get(id=m.id)
        assert not m.flag_a
        assert m.flag_b

        m.save()

        m = Bit1Model.objects.get(id=m.id)
        assert not m.flag_a
        assert m.flag_b

        m.flag_a = True
        m.flag_b = False
        m.save()

        m = Bit1Model.objects.get(id=m.id)
        assert m.flag_a
        assert not m.flag_b

    def test_defaults(self):
        m = Bit1Model.objects.create()
        assert m.flag_a
        assert not m.flag_b

        m = Bit1Model.objects.get(id=m.id)
        assert m.flag_a
        assert not m.flag_b

    def test_filter(self):
        m = Bit1Model.objects.create(flag_a=True, flag_b=True)

        assert list(Bit1Model.objects.filter(flag_a=True)) == [m]
        assert list(Bit1Model.objects.filter(flag_a=False)) == []

        assert list(Bit1Model.objects.filter(flag_a=F("flag_b"))) == [m]
        assert list(Bit1Model.objects.exclude(flag_a=F("flag_b"))) == []

        m.flag_a = False
        m.save()

        assert list(Bit1Model.objects.filter(flag_a=True)) == []
        assert list(Bit1Model.objects.filter(flag_a=False)) == [m]

        assert list(Bit1Model.objects.filter(flag_a=F("flag_b"))) == []
        assert list(Bit1Model.objects.exclude(flag_a=F("flag_b"))) == [m]

        Bit1Model.objects.filter(flag_a=False).update(flag_a=True)
        assert list(Bit1Model.objects.filter(flag_a=True)) == [m]
        assert list(Bit1Model.objects.filter(flag_a=False)) == []

        Bit1Model.objects.filter(flag_a=True).update(flag_a=False)
        assert list(Bit1Model.objects.filter(flag_a=True)) == []
        assert list(Bit1Model.objects.filter(flag_a=False)) == [m]


class TestSerialization(SimpleTestCase):
    def test_dumping(self):
        instance = Bit1Model(flag_a=True, flag_b=False)
        data = json.loads(serializers.serialize("json", [instance]))[0]
        fields = data["fields"]
        assert fields["flag_a"]
        assert not fields["flag_b"]

    def test_loading(self):
        test_data = """
            [{"fields": {"flag_a": false, "flag_b": true},
              "model": "testapp.Bit1Model", "pk": null}]
        """
        objs = list(serializers.deserialize("json", test_data))
        assert len(objs) == 1
        instance = objs[0].object
        assert not instance.flag_a
        assert instance.flag_b


class TestNullSaveLoad(TestCase):
    def test_basic(self):
        m = NullBit1Model()
        assert m.flag is None
        m.save()

        m = NullBit1Model.objects.get(id=m.id)
        print(m.flag)
        print(type(m.flag))
        assert m.flag is None

        m.flag = True
        m.save()

        m = NullBit1Model.objects.get(id=m.id)
        assert m.flag

        m.flag = False
        m.save()

        m = NullBit1Model.objects.get(id=m.id)
        assert m.flag is not None and not m.flag

    def test_defaults(self):
        m = NullBit1Model.objects.create()
        assert m.flag is None
        m = NullBit1Model.objects.get(id=m.id)
        assert m.flag is None

    def test_filter(self):
        m = NullBit1Model.objects.create()

        assert list(NullBit1Model.objects.filter(flag=None)) == [m]
        assert list(NullBit1Model.objects.filter(flag=True)) == []
        assert list(NullBit1Model.objects.filter(flag=False)) == []

        m.flag = True
        m.save()

        assert list(NullBit1Model.objects.filter(flag=None)) == []
        assert list(NullBit1Model.objects.filter(flag=True)) == [m]
        assert list(NullBit1Model.objects.filter(flag=False)) == []

        m.flag = False
        m.save()
        assert list(NullBit1Model.objects.filter(flag=None)) == []
        assert list(NullBit1Model.objects.filter(flag=True)) == []
        assert list(NullBit1Model.objects.filter(flag=False)) == [m]

        NullBit1Model.objects.filter(flag=False).update(flag=None)
        assert list(NullBit1Model.objects.filter(flag=None)) == [m]
        assert list(NullBit1Model.objects.filter(flag=True)) == []
        assert list(NullBit1Model.objects.filter(flag=False)) == []

        NullBit1Model.objects.filter(flag=None).update(flag=True)
        assert list(NullBit1Model.objects.filter(flag=None)) == []
        assert list(NullBit1Model.objects.filter(flag=True)) == [m]
        assert list(NullBit1Model.objects.filter(flag=False)) == []

        NullBit1Model.objects.filter(flag=True).update(flag=False)
        assert list(NullBit1Model.objects.filter(flag=None)) == []
        assert list(NullBit1Model.objects.filter(flag=True)) == []
        assert list(NullBit1Model.objects.filter(flag=False)) == [m]


class TestNullSerialization(SimpleTestCase):
    def test_dumping(self):
        instance = NullBit1Model(flag=None)
        data = json.loads(serializers.serialize("json", [instance]))[0]
        fields = data["fields"]
        assert fields["flag"] is None

    def test_loading(self):
        test_data = """
            [{"fields": {"flag": null},
              "model": "testapp.NullBit1Model", "pk": null}]
        """
        objs = list(serializers.deserialize("json", test_data))
        assert len(objs) == 1
        instance = objs[0].object
        assert instance.flag is None



from django.core import checks
from django.db.models import BinaryField, TextField


class SizedBinaryField(BinaryField):
    def __init__(self, *args, **kwargs):
        self.size_class = kwargs.pop("size_class", 4)
        super().__init__(*args, **kwargs)

    def check(self, **kwargs):
        errors = super().check(**kwargs)
        if self.size_class not in (1, 2, 3, 4):
            errors.append(
                checks.Error(
                    "size_class must be 1, 2, 3, or 4",
                    hint=None,
                    obj=self,
                    id="django_mysql.E007",
                )
            )
        return errors

    def deconstruct(self):
        name, path, args, kwargs = super().deconstruct()

        bad_paths = (
            "django_mysql.models.fields.sizes.SizedBinaryField",
            "django_mysql.models.fields.SizedBinaryField",
        )
        if path in bad_paths:
            path = "django_mysql.models.SizedBinaryField"

        kwargs["size_class"] = self.size_class
        return name, path, args, kwargs

    def db_type(self, connection):
        if self.size_class == 1:
            return "tinyblob"
        elif self.size_class == 2:
            return "blob"
        elif self.size_class == 3:
            return "mediumblob"
        else:  # don't check size_class == 4 as a safeguard for invalid values
            return "longblob"


class SizedTextField(TextField):
    def __init__(self, *args, **kwargs):
        self.size_class = kwargs.pop("size_class", 4)
        super().__init__(*args, **kwargs)

    def check(self, **kwargs):
        errors = super().check(**kwargs)
        if self.size_class not in (1, 2, 3, 4):
            errors.append(
                checks.Error(
                    "size_class must be 1, 2, 3, or 4",
                    hint=None,
                    obj=self,
                    id="django_mysql.E008",
                )
            )
        return errors

    def deconstruct(self):
        name, path, args, kwargs = super().deconstruct()

        bad_paths = (
            "django_mysql.models.fields.sizes.SizedTextField",
            "django_mysql.models.fields.SizedTextField",
        )
        if path in bad_paths:
            path = "django_mysql.models.SizedTextField"

        kwargs["size_class"] = self.size_class
        return name, path, args, kwargs

    def db_type(self, connection):
        if self.size_class == 1:
            return "tinytext"
        elif self.size_class == 2:
            return "text"
        elif self.size_class == 3:
            return "mediumtext"
        else:  # don't check size_class == 4 as a safeguard for invalid values
            return "longtext"



from io import StringIO
from unittest import mock

import pytest
from django.core.management import CommandError, call_command
from django.db.utils import ConnectionHandler
from django.test import SimpleTestCase

# Can't use @override_settings to swap out DATABASES, instead just mock.patch
# a new ConnectionHandler into the command module
command_connections = "django_mysql.management.commands.dbparams.connections"

sqlite = ConnectionHandler({"default": {"ENGINE": "django.db.backends.sqlite3"}})

full_db = ConnectionHandler(
    {
        "default": {
            "ENGINE": "django.db.backends.mysql",
            "NAME": "mydatabase",
            "USER": "ausername",
            "PASSWORD": "apassword",
            "HOST": "ahost.example.com",
            "PORT": "12345",
            "OPTIONS": {
                "read_default_file": "/tmp/defaults.cnf",
                "ssl": {"ca": "/tmp/mysql.cert"},
            },
        }
    }
)

socket_db = ConnectionHandler(
    {"default": {"ENGINE": "django.db.backends.mysql", "HOST": "/etc/mydb.sock"}}
)


class DBParamsTests(SimpleTestCase):
    def test_invalid_database(self):
        with pytest.raises(CommandError) as excinfo:
            call_command("dbparams", "nonexistent")
        assert "does not exist" in str(excinfo.value)

    def test_invalid_both(self):
        with pytest.raises(CommandError):
            call_command("dbparams", dsn=True, mysql=True)

    @mock.patch(command_connections, sqlite)
    def test_invalid_not_mysql(self):
        with pytest.raises(CommandError) as excinfo:
            call_command("dbparams")
        assert "not a MySQL database connection" in str(excinfo.value)

    @mock.patch(command_connections, full_db)
    def test_mysql_full(self):
        out = StringIO()
        call_command("dbparams", stdout=out)
        output = out.getvalue()
        assert output == (
            "--defaults-file=/tmp/defaults.cnf --user=ausername "
            + "--password=apassword --host=ahost.example.com --port=12345 "
            + "--ssl-ca=/tmp/mysql.cert mydatabase"
        )

    @mock.patch(command_connections, socket_db)
    def test_mysql_socket(self):
        out = StringIO()
        call_command("dbparams", stdout=out)
        output = out.getvalue()
        assert output == "--socket=/etc/mydb.sock"

    @mock.patch(command_connections, full_db)
    def test_dsn_full(self):
        out = StringIO()
        err = StringIO()
        call_command("dbparams", "default", dsn=True, stdout=out, stderr=err)
        output = out.getvalue()
        assert output == (
            "F=/tmp/defaults.cnf,u=ausername,p=apassword,"
            + "h=ahost.example.com,P=12345,D=mydatabase"
        )

        errors = err.getvalue()
        assert "SSL params can't be" in errors

    @mock.patch(command_connections, socket_db)
    def test_dsn_socket(self):
        out = StringIO()
        err = StringIO()
        call_command("dbparams", dsn=True, stdout=out, stderr=err)

        output = out.getvalue()
        assert output == "S=/etc/mydb.sock"

        errors = err.getvalue()
        assert errors == ""



from django.db import migrations

from django_mysql.models import EnumField


class Migration(migrations.Migration):

    dependencies = [("testapp", "0002_add_some_fields")]

    operations = [
        migrations.AlterField(
            model_name="EnumDefaultModel",
            name="field",
            field=EnumField(choices=[("lion", "Lion"), ("tiger", "Tiger"), "oh my!"]),
        )
    ]



from django.db.backends.mysql.base import DatabaseWrapper as BaseDatabaseWrapper


class DatabaseWrapper(BaseDatabaseWrapper):
    def init_connection_state(self):
        if not self.mysql_is_mariadb and self.mysql_version >= (8,):
            sql_mode = ",".join(
                [
                    "ERROR_FOR_DIVISION_BY_ZERO",
                    "NO_ZERO_DATE",
                    "NO_ZERO_IN_DATE",
                    "STRICT_TRANS_TABLES",
                ]
            )
        else:
            sql_mode = ",".join(
                [
                    "ERROR_FOR_DIVISION_BY_ZERO",
                    "NO_AUTO_CREATE_USER",
                    "NO_ZERO_DATE",
                    "NO_ZERO_IN_DATE",
                    "STRICT_TRANS_TABLES",
                ]
            )
        with self.cursor() as cursor:
            cursor.execute(
                """
                SET sql_mode=%s, innodb_strict_mode=1;
                """,
                (sql_mode,),
            )
            cursor.execute("SET NAMES 'utf8mb4' COLLATE 'utf8mb4_general_ci';")
        super().init_connection_state()



from unittest import SkipTest

import django
import pytest
from django.db import connection, migrations, models, transaction
from django.db.migrations.state import ProjectState
from django.test import TransactionTestCase
from django.test.utils import CaptureQueriesContext

from django_mysql.operations import AlterStorageEngine, InstallPlugin, InstallSOName
from django_mysql.test.utils import override_mysql_variables
from django_mysql.utils import connection_is_mariadb


def plugin_exists(plugin_name):
    with connection.cursor() as cursor:
        cursor.execute(
            """SELECT COUNT(*) FROM INFORMATION_SCHEMA.PLUGINS
               WHERE PLUGIN_NAME = %s""",
            (plugin_name,),
        )
        return cursor.fetchone()[0] > 0


def table_storage_engine(table_name):
    with connection.cursor() as cursor:
        cursor.execute(
            """SELECT ENGINE FROM INFORMATION_SCHEMA.TABLES
               WHERE TABLE_SCHEMA = DATABASE() AND TABLE_NAME = %s""",
            (table_name,),
        )
        return cursor.fetchone()[0]


class PluginOperationTests(TransactionTestCase):

    if django.VERSION >= (2, 2):
        databases = ["default", "other"]
    else:
        multi_db = True

    @classmethod
    def setUpClass(cls):
        super().setUpClass()
        has_metadata_lock_plugin = connection_is_mariadb(
            connection
        ) and connection.mysql_version >= (10, 0, 7)
        if not has_metadata_lock_plugin:
            raise SkipTest("The metadata_lock_info plugin is required")

    def test_install_plugin(self):
        """
        Test we can load the example plugin that every version of MySQL ships
        with.
        """
        assert not plugin_exists("metadata_lock_info")

        state = ProjectState()
        operation = InstallPlugin("metadata_lock_info", "metadata_lock_info.so")
        assert (
            operation.describe()
            == "Installs plugin metadata_lock_info from metadata_lock_info.so"
        )
        new_state = state.clone()
        with connection.schema_editor() as editor:
            operation.database_forwards("testapp", editor, state, new_state)

        assert plugin_exists("metadata_lock_info")

        new_state = state.clone()
        with connection.schema_editor() as editor:
            operation.database_backwards("testapp", editor, new_state, state)

        assert not plugin_exists("metadata_lock_info")

    def test_install_soname(self):
        """
        Test we can load the 'metadata_lock_info' library.
        """
        assert not plugin_exists("metadata_lock_info")

        state = ProjectState()
        operation = InstallSOName("metadata_lock_info.so")
        assert operation.describe() == "Installs library metadata_lock_info.so"

        new_state = state.clone()
        with connection.schema_editor() as editor:
            operation.database_forwards("testapp", editor, state, new_state)
        assert plugin_exists("metadata_lock_info")

        new_state = state.clone()
        with connection.schema_editor() as editor:
            operation.database_backwards("testapp", editor, new_state, state)
        assert not plugin_exists("metadata_lock_info")


class AlterStorageEngineTests(TransactionTestCase):
    def test_no_from_means_unreversible(self):
        operation = AlterStorageEngine("mymodel", to_engine="InnoDB")
        assert not operation.reversible

        with pytest.raises(NotImplementedError) as excinfo:
            operation.database_backwards(None, None, None, None)

        assert str(excinfo.value) == "You cannot reverse this operation"

    def test_from_means_reversible(self):
        operation = AlterStorageEngine(
            "mymodel", from_engine="MyISAM", to_engine="InnoDB"
        )
        assert operation.reversible

    def test_describe_without_from(self):
        operation = AlterStorageEngine("Pony", "InnoDB")
        assert operation.describe() == "Alter storage engine for Pony to InnoDB"

    def test_describe_with_from(self):
        operation = AlterStorageEngine("Pony", from_engine="MyISAM", to_engine="InnoDB")
        assert (
            operation.describe()
            == "Alter storage engine for Pony from MyISAM to InnoDB"
        )

    def test_references_model(self):
        operation = AlterStorageEngine("Pony", "InnoDB")
        assert operation.references_model("PONY")
        assert operation.references_model("Pony")
        assert operation.references_model("pony")
        assert not operation.references_model("poony")

    @override_mysql_variables(default_storage_engine="MyISAM")
    def test_running_with_changes(self):
        project_state = self.set_up_test_model("test_arstd")
        operation = AlterStorageEngine("Pony", from_engine="MyISAM", to_engine="InnoDB")

        assert table_storage_engine("test_arstd_pony") == "MyISAM"

        # Forwards
        new_state = project_state.clone()
        operation.state_forwards("test_arstd", new_state)
        with connection.schema_editor() as editor:
            operation.database_forwards("test_arstd", editor, project_state, new_state)
        assert table_storage_engine("test_arstd_pony") == "InnoDB"

        # Backwards
        with connection.schema_editor() as editor:
            operation.database_backwards("test_arstd", editor, new_state, project_state)
        assert table_storage_engine("test_arstd_pony") == "MyISAM"

    @override_mysql_variables(default_storage_engine="InnoDB")
    def test_running_without_changes(self):
        project_state = self.set_up_test_model("test_arstd")
        operation = AlterStorageEngine("Pony", from_engine="MyISAM", to_engine="InnoDB")

        assert table_storage_engine("test_arstd_pony") == "InnoDB"

        # Forwards - shouldn't actually do an ALTER since it is already InnoDB
        new_state = project_state.clone()
        operation.state_forwards("test_arstd", new_state)
        capturer = CaptureQueriesContext(connection)
        with capturer, connection.schema_editor() as editor:
            operation.database_forwards("test_arstd", editor, project_state, new_state)
        queries = [q["sql"] for q in capturer.captured_queries]
        assert not any(q.startswith("ALTER TABLE ") for q in queries)
        assert table_storage_engine("test_arstd_pony") == "InnoDB"

        # Backwards - will actually ALTER since it is going 'back' to MyISAM
        with connection.schema_editor() as editor:
            operation.database_backwards("test_arstd", editor, new_state, project_state)
        assert table_storage_engine("test_arstd_pony") == "MyISAM"

    # Copied from django core migration tests

    def set_up_test_model(
        self,
        app_label,
        second_model=False,
        third_model=False,
        related_model=False,
        mti_model=False,
        proxy_model=False,
        unique_together=False,
        options=False,
        db_table=None,
        index_together=False,
    ):
        """
        Creates a test model state and database table.
        """
        # Delete the tables if they already exist
        table_names = [
            # Start with ManyToMany tables
            "_pony_stables",
            "_pony_vans",
            # Then standard model tables
            "_pony",
            "_stable",
            "_van",
        ]
        tables = [(app_label + table_name) for table_name in table_names]
        with connection.cursor() as cursor:
            table_names = connection.introspection.table_names(cursor)
            connection.disable_constraint_checking()
            sql_delete_table = connection.schema_editor().sql_delete_table
            with transaction.atomic():
                for table in tables:
                    if table in table_names:
                        cursor.execute(
                            sql_delete_table
                            % {"table": connection.ops.quote_name(table)}
                        )
            connection.enable_constraint_checking()

        # Make the "current" state
        model_options = {
            "swappable": "TEST_SWAP_MODEL",
            "index_together": [["weight", "pink"]] if index_together else [],
            "unique_together": [["pink", "weight"]] if unique_together else [],
        }
        if options:
            model_options["permissions"] = [("can_groom", "Can groom")]
        if db_table:
            model_options["db_table"] = db_table
        operations = [
            migrations.CreateModel(
                "Pony",
                [
                    ("id", models.AutoField(primary_key=True)),
                    ("pink", models.IntegerField(default=3)),
                    ("weight", models.FloatField()),
                ],
                options=model_options,
            )
        ]
        if second_model:
            operations.append(
                migrations.CreateModel(
                    "Stable", [("id", models.AutoField(primary_key=True))]
                )
            )
        if third_model:
            operations.append(
                migrations.CreateModel(
                    "Van", [("id", models.AutoField(primary_key=True))]
                )
            )
        if related_model:
            operations.append(
                migrations.CreateModel(
                    "Rider",
                    [
                        ("id", models.AutoField(primary_key=True)),
                        ("pony", models.ForeignKey("Pony")),
                        ("friend", models.ForeignKey("self")),
                    ],
                )
            )
        if mti_model:
            operations.append(
                migrations.CreateModel(
                    "ShetlandPony",
                    fields=[
                        (
                            "pony_ptr",
                            models.OneToOneField(
                                auto_created=True,
                                primary_key=True,
                                to_field="id",
                                serialize=False,
                                to="Pony",
                            ),
                        ),
                        ("cuteness", models.IntegerField(default=1)),
                    ],
                    bases=["%s.Pony" % app_label],
                )
            )
        if proxy_model:
            operations.append(
                migrations.CreateModel(
                    "ProxyPony",
                    fields=[],
                    options={"proxy": True},
                    bases=["%s.Pony" % app_label],
                )
            )

        return self.apply_operations(app_label, ProjectState(), operations)

    def apply_operations(self, app_label, project_state, operations):
        migration = migrations.Migration("name", app_label)
        migration.operations = operations
        with connection.schema_editor() as editor:
            return migration.apply(project_state, editor)



from django.core.validators import MaxLengthValidator, MinLengthValidator
from django.utils.translation import ungettext_lazy


class ListMaxLengthValidator(MaxLengthValidator):
    message = ungettext_lazy(
        "List contains %(show_value)d item, "
        "it should contain no more than %(limit_value)d.",
        "List contains %(show_value)d items, "
        "it should contain no more than %(limit_value)d.",
        "limit_value",
    )


class ListMinLengthValidator(MinLengthValidator):
    message = ungettext_lazy(
        "List contains %(show_value)d item, "
        "it should contain no fewer than %(limit_value)d.",
        "List contains %(show_value)d items, "
        "it should contain no fewer than %(limit_value)d.",
        "limit_value",
    )


class SetMaxLengthValidator(MaxLengthValidator):
    message = ungettext_lazy(
        "Set contains %(show_value)d item, "
        "it should contain no more than %(limit_value)d.",
        "Set contains %(show_value)d items, "
        "it should contain no more than %(limit_value)d.",
        "limit_value",
    )


class SetMinLengthValidator(MinLengthValidator):
    message = ungettext_lazy(
        "Set contains %(show_value)d item, "
        "it should contain no fewer than %(limit_value)d.",
        "Set contains %(show_value)d items, "
        "it should contain no fewer than %(limit_value)d.",
        "limit_value",
    )



import json

from django import forms
from django.core import validators
from django.core.exceptions import ValidationError
from django.utils.text import format_lazy
from django.utils.translation import ugettext_lazy as _

from django_mysql.validators import (
    ListMaxLengthValidator,
    ListMinLengthValidator,
    SetMaxLengthValidator,
    SetMinLengthValidator,
)


class SimpleListField(forms.CharField):

    default_error_messages = {
        "item_n_invalid": _("Item %(nth)s in the list did not validate: "),
        "no_double_commas": _("No leading, trailing, or double commas."),
    }

    def __init__(self, base_field, max_length=None, min_length=None, *args, **kwargs):
        self.base_field = base_field
        super().__init__(*args, **kwargs)
        if max_length is not None:
            self.max_length = max_length
            self.validators.append(ListMaxLengthValidator(int(max_length)))
        if min_length is not None:
            self.min_length = min_length
            self.validators.append(ListMinLengthValidator(int(min_length)))

    def prepare_value(self, value):
        if isinstance(value, list):
            return ",".join(str(self.base_field.prepare_value(v)) for v in value)
        return value

    def to_python(self, value):
        if value and len(value):
            items = value.split(",")
        else:
            items = []

        errors = []
        values = []
        for i, item in enumerate(items, start=1):
            if not len(item):
                errors.append(
                    ValidationError(
                        self.error_messages["no_double_commas"], code="no_double_commas"
                    )
                )
                continue

            try:
                value = self.base_field.to_python(item)
            except ValidationError as e:
                for error in e.error_list:
                    errors.append(
                        ValidationError(
                            format_lazy(
                                "{}{}",
                                self.error_messages["item_n_invalid"],
                                error.message,
                            ),
                            code="item_n_invalid",
                            params={"nth": i},
                        )
                    )

            values.append(value)

        if errors:
            raise ValidationError(errors)

        return values

    def validate(self, value):
        super().validate(value)
        errors = []
        for i, item in enumerate(value, start=1):
            try:
                self.base_field.validate(item)
            except ValidationError as e:
                for error in e.error_list:
                    for message in error.messages:
                        errors.append(
                            ValidationError(
                                format_lazy(
                                    "{}{}",
                                    self.error_messages["item_n_invalid"],
                                    message,
                                ),
                                code="item_invalid",
                                params={"nth": i},
                            )
                        )
        if errors:
            raise ValidationError(errors)

    def run_validators(self, value):
        super().run_validators(value)
        errors = []
        for i, item in enumerate(value, start=1):
            try:
                self.base_field.run_validators(item)
            except ValidationError as e:
                for error in e.error_list:
                    for message in error.messages:
                        errors.append(
                            ValidationError(
                                format_lazy(
                                    "{}{}",
                                    self.error_messages["item_n_invalid"],
                                    message,
                                ),
                                code="item_n_invalid",
                                params={"nth": i},
                            )
                        )
        if errors:
            raise ValidationError(errors)


class SimpleSetField(forms.CharField):
    empty_values = list(validators.EMPTY_VALUES) + [set()]

    default_error_messages = {
        "item_invalid": _('Item "%(item)s" in the set did not validate: '),
        "item_n_invalid": _("Item %(nth)s in the set did not validate: "),
        "no_double_commas": _("No leading, trailing, or double commas."),
        "no_duplicates": _(
            "Duplicates are not supported. " "'%(item)s' appears twice or more."
        ),
    }

    def __init__(self, base_field, max_length=None, min_length=None, *args, **kwargs):
        self.base_field = base_field
        super().__init__(*args, **kwargs)
        if max_length is not None:
            self.max_length = max_length
            self.validators.append(SetMaxLengthValidator(int(max_length)))
        if min_length is not None:
            self.min_length = min_length
            self.validators.append(SetMinLengthValidator(int(min_length)))

    def prepare_value(self, value):
        if isinstance(value, set):
            return ",".join(str(self.base_field.prepare_value(v)) for v in value)
        return value

    def to_python(self, value):
        if value and len(value):
            items = value.split(",")
        else:
            items = []

        errors = []
        values = set()
        for i, item in enumerate(items, start=1):
            if not len(item):
                errors.append(
                    ValidationError(
                        self.error_messages["no_double_commas"], code="no_double_commas"
                    )
                )
                continue

            try:
                value = self.base_field.to_python(item)
            except ValidationError as e:
                for error in e.error_list:
                    errors.append(
                        ValidationError(
                            format_lazy(
                                "{}{}",
                                self.error_messages["item_n_invalid"],
                                error.message,
                            ),
                            code="item_n_invalid",
                            params={"nth": i},
                        )
                    )

            if value in values:
                errors.append(
                    ValidationError(
                        self.error_messages["no_duplicates"],
                        code="no_duplicates",
                        params={"item": item},
                    )
                )
            else:
                values.add(value)

        if errors:
            raise ValidationError(errors)

        return values

    def validate(self, value):
        super().validate(value)
        errors = []
        for item in value:
            try:
                self.base_field.validate(item)
            except ValidationError as e:
                for error in e.error_list:
                    for message in error.messages:
                        errors.append(
                            ValidationError(
                                format_lazy(
                                    "{}{}", self.error_messages["item_invalid"], message
                                ),
                                code="item_invalid",
                                params={"item": item},
                            )
                        )
        if errors:
            raise ValidationError(errors)

    def run_validators(self, value):
        super().run_validators(value)
        errors = []
        for item in value:
            try:
                self.base_field.run_validators(item)
            except ValidationError as e:
                for error in e.error_list:
                    for message in error.messages:
                        errors.append(
                            ValidationError(
                                format_lazy(
                                    "{}{}", self.error_messages["item_invalid"], message
                                ),
                                code="item_invalid",
                                params={"item": item},
                            )
                        )
        if errors:
            raise ValidationError(errors)


class InvalidJSONInput(str):
    pass


class JSONString(str):
    pass


class JSONField(forms.CharField):
    default_error_messages = {"invalid": _("'%(value)s' value must be valid JSON.")}
    widget = forms.Textarea

    def to_python(self, value):
        if self.disabled:
            return value
        if value in self.empty_values:
            return None
        elif isinstance(value, (list, dict, int, float, JSONString)):
            return value
        try:
            converted = json.loads(value)
        except ValueError:
            raise forms.ValidationError(
                self.error_messages["invalid"], code="invalid", params={"value": value}
            )
        if isinstance(converted, str):
            return JSONString(converted)
        else:
            return converted

    def bound_data(self, data, initial):
        if self.disabled:
            return initial
        try:
            return json.loads(data)
        except ValueError:
            return InvalidJSONInput(data)

    def prepare_value(self, value):
        if isinstance(value, InvalidJSONInput):
            return value
        return json.dumps(value)



import pytest
from django.core.management import call_command
from django.db import connection
from django.db.migrations.writer import MigrationWriter
from django.db.transaction import atomic
from django.db.utils import DataError
from django.test import TestCase, TransactionTestCase
from django.test.utils import override_settings

from django_mysql.models import SizedBinaryField, SizedTextField
from django_mysql.test.utils import override_mysql_variables
from tests.testapp.models import SizeFieldModel, TemporaryModel
from tests.testapp.utils import column_type

# Ensure we aren't just warned about the data truncation
forceDataError = override_mysql_variables(SQL_MODE="STRICT_TRANS_TABLES")


def migrate(name):
    call_command(
        "migrate", "testapp", name, verbosity=0, skip_checks=True, interactive=False
    )


class SubSizedBinaryField(SizedBinaryField):
    """
    Used below, has a different path for deconstruct()
    """


@forceDataError
class SizedBinaryFieldTests(TestCase):
    def test_binaryfield_checks(self):
        class InvalidSizedBinaryModel(TemporaryModel):
            field = SizedBinaryField(size_class=5)

        errors = InvalidSizedBinaryModel.check(actually_check=True)
        assert len(errors) == 1
        assert errors[0].id == "django_mysql.E007"
        assert errors[0].msg == "size_class must be 1, 2, 3, or 4"

    def test_binaryfield_default_length(self):
        # By default, SizedBinaryField should act like BinaryField
        field = SizedBinaryField()
        assert field.size_class == 4
        assert field.db_type(None) == "longblob"

    @atomic
    def test_binary_1_max_length(self):
        # Okay
        m = SizeFieldModel(binary1=bytes(1) * (2 ** 8 - 1))
        m.save()

        # Bad - Data too long
        m = SizeFieldModel(binary1=bytes(1) * (2 ** 8))
        with pytest.raises(DataError) as excinfo:
            m.save()
        assert excinfo.value.args[0] == 1406

    def test_deconstruct_path(self):
        field = SizedBinaryField(size_class=1)
        name, path, args, kwargs = field.deconstruct()
        assert path == "django_mysql.models.SizedBinaryField"

    def test_deconstruct_subclass_path(self):
        field = SubSizedBinaryField(size_class=1)
        name, path, args, kwargs = field.deconstruct()
        assert path == "tests.testapp.test_size_fields.SubSizedBinaryField"

    def test_deconstruct_size_class_4(self):
        field = SizedBinaryField(size_class=4)
        name, path, args, kwargs = field.deconstruct()
        new = SizedBinaryField(*args, **kwargs)
        assert new.size_class == field.size_class

    def test_deconstruct_size_class_2(self):
        field = SizedBinaryField(size_class=2)
        name, path, args, kwargs = field.deconstruct()
        new = SizedBinaryField(*args, **kwargs)
        assert new.size_class == field.size_class

    def test_makemigrations(self):
        field = SizedBinaryField(size_class=1)
        statement, imports = MigrationWriter.serialize(field)

        assert statement == "django_mysql.models.SizedBinaryField(size_class=1)"

    def test_makemigrations_size_class_implicit(self):
        field = SizedBinaryField()
        statement, imports = MigrationWriter.serialize(field)

        assert statement == "django_mysql.models.SizedBinaryField(size_class=4)"


@forceDataError
class SizedBinaryFieldMigrationTests(TransactionTestCase):
    @override_settings(
        MIGRATION_MODULES={"testapp": "tests.testapp.sizedbinaryfield_migrations"}
    )
    def test_adding_field_with_default(self):
        table_name = "testapp_sizedbinaryaltermodel"
        table_names = connection.introspection.table_names

        with connection.cursor() as cursor:
            assert table_name not in table_names(cursor)

        migrate("0001_initial")
        with connection.cursor() as cursor:
            assert table_name in table_names(cursor)
            assert column_type(table_name, "field") == "longblob"

        migrate("0002_alter_field")
        with connection.cursor() as cursor:
            assert table_name in table_names(cursor)
            assert column_type(table_name, "field") == "blob"

        migrate("0001_initial")
        with connection.cursor() as cursor:
            assert table_name in table_names(cursor)
            assert column_type(table_name, "field") == "longblob"

        migrate("zero")
        with connection.cursor() as cursor:
            assert table_name not in table_names(cursor)


class SubSizedTextField(SizedTextField):
    """
    Used below, has a different path for deconstruct()
    """


@forceDataError
class SizedTextFieldTests(TestCase):
    def test_check_max_length(self):
        class InvalidSizedTextModel(TemporaryModel):
            field = SizedTextField(size_class=5)

        errors = InvalidSizedTextModel.check(actually_check=True)
        assert len(errors) == 1
        assert errors[0].id == "django_mysql.E008"
        assert errors[0].msg == "size_class must be 1, 2, 3, or 4"

    def test_textfield_default_length(self):
        # By default, SizedTextField should act like TextField
        field = SizedTextField()
        assert field.size_class == 4
        assert field.db_type(None) == "longtext"

    def test_tinytext_max_length(self):
        # Okay
        m = SizeFieldModel(text1="a" * (2 ** 8 - 1))
        m.save()

        # Bad - Data too long
        m = SizeFieldModel(text1="a" * (2 ** 8))
        with atomic(), pytest.raises(DataError) as excinfo:
            m.save()
        assert excinfo.value.args[0] == 1406

    def test_deconstruct_path(self):
        field = SizedTextField(size_class=1)
        name, path, args, kwargs = field.deconstruct()
        assert path == "django_mysql.models.SizedTextField"

    def test_deconstruct_subclass_path(self):
        field = SubSizedTextField(size_class=1)
        name, path, args, kwargs = field.deconstruct()
        assert path == "tests.testapp.test_size_fields.SubSizedTextField"

    def test_deconstruct_size_class_4(self):
        field = SizedTextField(size_class=4)
        name, path, args, kwargs = field.deconstruct()
        new = SizedTextField(*args, **kwargs)
        assert new.size_class == field.size_class

    def test_deconstruct_size_class_2(self):
        field = SizedTextField(size_class=2)
        name, path, args, kwargs = field.deconstruct()
        new = SizedTextField(*args, **kwargs)
        assert new.size_class == field.size_class

    def test_makemigrations(self):
        field = SizedTextField(size_class=1)
        statement, imports = MigrationWriter.serialize(field)

        assert statement == "django_mysql.models.SizedTextField(size_class=1)"

    def test_makemigrations_size_class_implicit(self):
        field = SizedTextField()
        statement, imports = MigrationWriter.serialize(field)

        assert statement == "django_mysql.models.SizedTextField(size_class=4)"


@forceDataError
class SizedTextFieldMigrationTests(TransactionTestCase):
    @override_settings(
        MIGRATION_MODULES={"testapp": "tests.testapp.sizedtextfield_migrations"}
    )
    def test_adding_field_with_default(self):
        table_name = "testapp_sizedtextaltermodel"
        table_names = connection.introspection.table_names

        with connection.cursor() as cursor:
            assert table_name not in table_names(cursor)

        migrate("0001_initial")
        with connection.cursor() as cursor:
            assert table_name in table_names(cursor)
            assert column_type(table_name, "field") == "mediumtext"

        migrate("0002_alter_field")
        with connection.cursor() as cursor:
            assert table_name in table_names(cursor)
            assert column_type(table_name, "field") == "tinytext"

        migrate("0001_initial")
        with connection.cursor() as cursor:
            assert table_name in table_names(cursor)
            assert column_type(table_name, "field") == "mediumtext"

        migrate("zero")
        with connection.cursor() as cursor:
            assert table_name not in table_names(cursor)



from django.conf import settings
from django.core.cache import InvalidCacheBackendError, caches
from django.core.management import BaseCommand, CommandError

from django_mysql.cache import MySQLCache
from django_mysql.utils import collapse_spaces


class Command(BaseCommand):
    args = "<optional cache aliases>"

    help = collapse_spaces(
        """
        Runs cache.cull() on all your MySQLCache caches, or only those
        specified aliases.
    """
    )

    def add_arguments(self, parser):
        parser.add_argument(
            "aliases",
            metavar="aliases",
            nargs="*",
            help="Specify the cache alias(es) to cull.",
        )

    def handle(self, *args, **options):
        verbosity = options.get("verbosity")

        aliases = set(options["aliases"])

        if not aliases:
            aliases = settings.CACHES

        for alias in aliases:
            try:
                cache = caches[alias]
            except InvalidCacheBackendError:
                raise CommandError("Cache '{}' does not exist".format(alias))

            if not isinstance(cache, MySQLCache):  # pragma: no cover
                continue

            if verbosity >= 1:
                self.stdout.write(
                    "Deleting from cache '{}'... ".format(alias), ending=""
                )
            num_deleted = cache.cull()
            if verbosity >= 1:
                self.stdout.write("{} entries deleted.".format(num_deleted))



import json
from decimal import Decimal
from unittest import SkipTest, mock

import django
import pytest
from django.core import serializers
from django.core.serializers.json import DjangoJSONEncoder
from django.db import connection
from django.db.models import F
from django.test import TestCase

from django_mysql import forms
from django_mysql.models import JSONField
from django_mysql.utils import connection_is_mariadb
from tests.testapp.models import JSONModel, TemporaryModel
from tests.testapp.utils import mock_mysql_version


class JSONFieldTestCase(TestCase):
    @classmethod
    def setUpClass(cls):
        if not (
            not connection_is_mariadb(connection) and connection.mysql_version >= (5, 7)
        ):
            raise SkipTest("JSONField requires MySQL 5.7+")
        super().setUpClass()


class TestSaveLoad(JSONFieldTestCase):
    def test_empty_dict(self):
        m = JSONModel()
        assert m.attrs == {}
        m.save()
        m = JSONModel.objects.get()
        assert m.attrs == {}

    def test_values(self):
        m = JSONModel(attrs={"key": "value"})
        assert m.attrs == {"key": "value"}
        m.save()
        m = JSONModel.objects.get()
        assert m.attrs == {"key": "value"}

    def test_string(self):
        m = JSONModel(attrs="value")
        assert m.attrs == "value"
        m.save()
        m = JSONModel.objects.get()
        assert m.attrs == "value"

    def test_json_dumps_string(self):
        json_string = json.dumps({"foo": "bar"})
        m = JSONModel(attrs=json_string)
        assert m.attrs == json_string
        m.save()
        m = JSONModel.objects.get()
        assert m.attrs == json_string

    def test_awkward_1(self):
        m = JSONModel(attrs='"')
        assert m.attrs == '"'
        m.save()
        m = JSONModel.objects.get()
        assert m.attrs == '"'

    def test_awkward_2(self):
        m = JSONModel(attrs="\\")
        assert m.attrs == "\\"
        m.save()
        m = JSONModel.objects.get()
        assert m.attrs == "\\"

    def test_list(self):
        m = JSONModel(attrs=[1, 2, 4])
        assert m.attrs == [1, 2, 4]
        m.save()
        m = JSONModel.objects.get()
        assert m.attrs == [1, 2, 4]

    def test_true(self):
        m = JSONModel(attrs=True)
        assert m.attrs is True
        m.save()
        m = JSONModel.objects.get()
        assert m.attrs is True

    def test_false(self):
        m = JSONModel(attrs=False)
        assert m.attrs is False
        m.save()
        m = JSONModel.objects.get()
        assert m.attrs is False

    def test_null(self):
        m = JSONModel(attrs=None)
        assert m.attrs is None
        m.save()
        m = JSONModel.objects.get()
        assert m.attrs is None

    def test_control_characters(self):
        chars = "".join(chr(i) for i in range(32))
        m = JSONModel(attrs=[chars])
        assert m.attrs == [chars]
        m.save()
        m = JSONModel.objects.get()
        assert m.attrs == [chars]

    def test_nan_raises_valueerror(self):
        m = JSONModel(attrs=float("nan"))
        with pytest.raises(ValueError):
            m.save()

    def test_custom_json_encoder(self):
        field = JSONField(encoder=DjangoJSONEncoder(allow_nan=False))
        value = field.get_prep_value({"a": Decimal(1)})
        assert value == '{"a": "1"}'

    def test_custom_json_decoder(self):
        class CustomDecoder(json.JSONDecoder):
            def decode(self, *args, **kwargs):
                return "lol"

        field = JSONField(decoder=CustomDecoder(strict=False))
        value = field.from_db_value('"anything"', None, None)
        assert value == "lol"


class QueryTests(JSONFieldTestCase):
    def setUp(self):
        super().setUp()
        JSONModel.objects.bulk_create(
            [
                JSONModel(attrs={"a": "b"}, name="b"),
                JSONModel(attrs=1337, name="x"),
                JSONModel(attrs=["an", "array"]),
                JSONModel(attrs=None),
                JSONModel(attrs="foo", name="foo"),
                JSONModel(attrs=True),
                JSONModel(attrs=False),
                JSONModel(attrs=0.5),
            ]
        )
        self.objs = list(JSONModel.objects.all().order_by("id"))

    def test_equal(self):
        assert list(JSONModel.objects.filter(attrs={"a": "b"})) == [self.objs[0]]

    def test_equal_int(self):
        assert list(JSONModel.objects.filter(attrs=1337)) == [self.objs[1]]

    def test_equal_float(self):
        assert list(JSONModel.objects.filter(attrs=0.5)) == [self.objs[7]]
        assert list(JSONModel.objects.filter(attrs=0.501)) == []

    def test_equal_string(self):
        assert list(JSONModel.objects.filter(attrs="foo")) == [self.objs[4]]

    def test_equal_array(self):
        assert list(JSONModel.objects.filter(attrs=["an", "array"])) == [self.objs[2]]

    def test_equal_bool(self):
        assert list(JSONModel.objects.filter(attrs=True)) == [self.objs[5]]
        assert list(JSONModel.objects.filter(attrs=False)) == [self.objs[6]]

    def test_equal_no_match(self):
        assert list(JSONModel.objects.filter(attrs={"c": "z"})) == []

    def test_equal_F_attrs(self):
        assert list(JSONModel.objects.filter(attrs=F("attrs"))) == [
            self.objs[0],
            self.objs[1],
            self.objs[2],
            self.objs[4],
            self.objs[5],
            self.objs[6],
            self.objs[7],
        ]
        assert list(JSONModel.objects.filter(attrs=F("name"))) == [self.objs[4]]
        assert list(JSONModel.objects.filter(attrs__a=F("name"))) == [self.objs[0]]

    def test_isnull_True(self):
        assert list(JSONModel.objects.filter(attrs__isnull=True)) == [self.objs[3]]

    def test_isnull_False(self):
        assert list(JSONModel.objects.filter(attrs__isnull=False)) == [
            self.objs[0],
            self.objs[1],
            self.objs[2],
            self.objs[4],
            self.objs[5],
            self.objs[6],
            self.objs[7],
        ]

    def test_range_broken(self):
        with pytest.raises(NotImplementedError) as excinfo:
            JSONModel.objects.filter(attrs__range=[1, 2])

        assert "Lookup 'range' doesn't work with JSONField" in str(excinfo.value)

    def test_in_lookup(self):
        objs = list(
            JSONModel.objects.filter(attrs__in=[{"a": "c"}, {"a": "b"}, {"c": "b"}])
        )
        assert objs == [self.objs[0]]

        objs = list(JSONModel.objects.filter(attrs__a__in=["b"]))
        assert objs == [self.objs[0]]

        objs = list(JSONModel.objects.filter(attrs__in=[111, 1337, 333]))
        assert objs == [self.objs[1]]

        objs = list(
            JSONModel.objects.filter(
                attrs__in=[["array"], ["an", "array"], ["another", "array"]]
            )
        )
        assert objs == [self.objs[2]]

        objs = list(JSONModel.objects.filter(attrs__in=["bar", "foo", "baz"]))
        assert objs == [self.objs[4]]

        objs = list(
            JSONModel.objects.filter(
                attrs__in=[{"a": "b"}, 1337, "foo", ["an", "array"]]
            )
        )
        assert objs == [self.objs[0], self.objs[1], self.objs[2], self.objs[4]]

        objs = list(JSONModel.objects.filter(attrs__in=["bar"]))
        assert objs == []


class ArrayQueryTests(JSONFieldTestCase):
    def setUp(self):
        super().setUp()
        JSONModel.objects.bulk_create(
            [
                JSONModel(attrs=[1, 3]),
                JSONModel(attrs=[1, 3, 3]),
                JSONModel(attrs=[1, 3, 3, 7]),
                JSONModel(attrs=[2, 4]),
            ]
        )
        self.objs = list(JSONModel.objects.all().order_by("id"))

    def test_lt_1(self):
        assert list(JSONModel.objects.filter(attrs__lt=[1])) == []

    def test_lt_3(self):
        assert list(JSONModel.objects.filter(attrs__lt=[3])) == self.objs

    def test_lte_1_3_3(self):
        assert list(JSONModel.objects.filter(attrs__lte=[1, 3, 3])) == [
            self.objs[0],
            self.objs[1],
        ]

    def test_lte_1(self):
        assert list(JSONModel.objects.filter(attrs__lte=[1])) == []

    def test_gt_1(self):
        assert list(JSONModel.objects.filter(attrs__gt=[1])) == self.objs

    def test_gt_1_3(self):
        assert list(JSONModel.objects.filter(attrs__gt=[1, 3])) == self.objs[1:]

    def test_gt_2_5(self):
        assert list(JSONModel.objects.filter(attrs__gt=[2, 5])) == []

    def test_gte_1_3(self):
        assert list(JSONModel.objects.filter(attrs__gte=[1, 3])) == self.objs

    def test_gte_2(self):
        assert list(JSONModel.objects.filter(attrs__gte=[2])) == [self.objs[3]]


class ExtraLookupsQueryTests(JSONFieldTestCase):
    def setUp(self):
        super().setUp()

        self.objs = [
            JSONModel.objects.create(attrs={}),
            JSONModel.objects.create(
                name='"b"',
                attrs={
                    "a": "b",
                    "c": 1,
                    9001: 9002,
                    '"': "awkward",
                    "\n": "super awkward",
                },
            ),
            JSONModel.objects.create(
                name="b",
                attrs={
                    "a": "b",
                    "c": 1,
                    "d": ["e", {"f": "g"}, 1],
                    "h": True,
                    "i": False,
                    "j": None,
                    "k": {"l": "m"},
                },
            ),
            JSONModel.objects.create(attrs=[1, [2]]),
            JSONModel.objects.create(attrs={"k": True, "l": False, "\\": "awkward"}),
            JSONModel.objects.create(attrs={"a": "n"}, name="x"),
        ]

    def test_has_key_invalid_type(self):
        with pytest.raises(ValueError) as excinfo:
            JSONModel.objects.filter(attrs__has_key=1)
        assert (
            str(excinfo.value)
            == "JSONField's 'has_key' lookup only works with str values"
        )

    def test_has_key(self):
        assert list(JSONModel.objects.filter(attrs__has_key="a")) == [
            self.objs[1],
            self.objs[2],
            self.objs[5],
        ]

    def test_has_key_2(self):
        assert list(JSONModel.objects.filter(attrs__has_key="l")) == [self.objs[4]]

    def test_has_key_awkward(self):
        assert list(JSONModel.objects.filter(attrs__has_key='"')) == [self.objs[1]]

    def test_has_key_awkward_2(self):
        assert list(JSONModel.objects.filter(attrs__has_key="\n")) == [self.objs[1]]

    def test_has_key_awkward_3(self):
        assert list(JSONModel.objects.filter(attrs__has_key="\\")) == [self.objs[4]]

    def test_has_keys_invalid_type(self):
        with pytest.raises(ValueError) as excinfo:
            JSONModel.objects.filter(attrs__has_keys={"a": "b"})
        assert "only works with Sequences" in str(excinfo.value)

    def test_has_keys(self):
        assert list(JSONModel.objects.filter(attrs__has_keys=["a", "c"])) == [
            self.objs[1],
            self.objs[2],
        ]

    def test_has_keys_2(self):
        assert list(JSONModel.objects.filter(attrs__has_keys=["l"])) == [self.objs[4]]

    def test_has_keys_awkward(self):
        assert list(JSONModel.objects.filter(attrs__has_keys=["\n", '"'])) == [
            self.objs[1]
        ]

    def test_has_any_keys_invalid_type(self):
        with pytest.raises(ValueError) as excinfo:
            JSONModel.objects.filter(attrs__has_any_keys={"a": "b"})
        assert "only works with Sequences" in str(excinfo.value)

    def test_has_any_keys(self):
        assert list(JSONModel.objects.filter(attrs__has_any_keys=["a", "k"])) == [
            self.objs[1],
            self.objs[2],
            self.objs[4],
            self.objs[5],
        ]

    def test_has_any_keys_awkward(self):
        assert list(JSONModel.objects.filter(attrs__has_any_keys=["\\"])) == [
            self.objs[4]
        ]

    def test_contains_scalar_in_scalar(self):
        assert list(JSONModel.objects.filter(attrs__a__contains="b")) == [
            self.objs[1],
            self.objs[2],
        ]
        assert list(JSONModel.objects.filter(attrs__c__contains=1)) == [
            self.objs[1],
            self.objs[2],
        ]
        assert list(JSONModel.objects.filter(attrs__l__contains=False)) == [
            self.objs[4],
        ]
        assert list(JSONModel.objects.filter(attrs__h__contains=True)) == [
            self.objs[2],
        ]
        assert list(JSONModel.objects.filter(attrs__j__contains=None)) == [
            self.objs[2],
        ]

        assert list(JSONModel.objects.filter(attrs__a__contains=1)) == []
        assert list(JSONModel.objects.filter(attrs__c__contains="b")) == []
        assert list(JSONModel.objects.filter(attrs__l__contains=True)) == []
        assert list(JSONModel.objects.filter(attrs__h__contains=False)) == []

    def test_contains_object_in_object(self):
        assert list(JSONModel.objects.filter(attrs__contains={"a": "b"})) == [
            self.objs[1],
            self.objs[2],
        ]
        assert list(JSONModel.objects.filter(attrs__contains={"c": 1})) == [
            self.objs[1],
            self.objs[2],
        ]
        assert list(JSONModel.objects.filter(attrs__contains={"l": False})) == [
            self.objs[4],
        ]
        assert list(JSONModel.objects.filter(attrs__contains={"h": True})) == [
            self.objs[2],
        ]
        assert list(JSONModel.objects.filter(attrs__contains={"j": None})) == [
            self.objs[2],
        ]

        assert list(JSONModel.objects.filter(attrs__contains={"a": 1})) == []
        assert list(JSONModel.objects.filter(attrs__contains={"c": "b"})) == []
        assert list(JSONModel.objects.filter(attrs__contains={"l": True})) == []
        assert list(JSONModel.objects.filter(attrs__contains={"h": False})) == []

    def test_contains_array(self):
        assert list(JSONModel.objects.filter(attrs__contains=[[2]])) == [self.objs[3]]
        assert list(JSONModel.objects.filter(attrs__contains=1)) == [self.objs[3]]
        assert list(JSONModel.objects.filter(attrs__d__contains="e")) == [self.objs[2]]
        assert list(JSONModel.objects.filter(attrs__d__contains=1)) == [self.objs[2]]
        assert list(JSONModel.objects.filter(attrs__d__contains={"f": "g"})) == [
            self.objs[2]
        ]
        assert list(JSONModel.objects.filter(attrs__d__contains=2)) == []

    def test_contains_expression_and_subqueries(self):
        with pytest.raises(ValueError) as excinfo:
            list(JSONModel.objects.filter(attrs__a__contains=F("name")))
        assert (
            str(excinfo.value)
            == "JSONField's 'contains' lookup only works with literal values"
        )

        subquery = JSONModel.objects.filter(name="b").values_list("attrs")
        with pytest.raises(ValueError) as excinfo:
            JSONModel.objects.filter(attrs__contains=subquery)
        assert (
            str(excinfo.value)
            == "JSONField's 'contains' lookup only works with literal values"
        )

    def test_contained_by(self):
        assert list(
            JSONModel.objects.filter(
                attrs__contained_by={
                    "k": True,
                    "l": False,
                    "\\": "awkward",
                    "spare_key": ["unused", "array"],
                }
            )
        ) == [self.objs[0], self.objs[4]]

    def test_contained_by_array(self):
        assert list(JSONModel.objects.filter(attrs__contained_by=[1, [2], 3])) == [
            self.objs[3]
        ]

    def test_length_lookup(self):
        assert list(JSONModel.objects.filter(attrs__length=0)) == [self.objs[0]]

    def test_length_lookup_2(self):
        assert list(JSONModel.objects.filter(attrs__length=2)) == [self.objs[3]]

    def test_length_lookup_chains(self):
        assert list(JSONModel.objects.filter(attrs__length__range=[0, 10])) == self.objs

    def test_shallow_list_lookup(self):
        assert list(JSONModel.objects.filter(attrs__0=1)) == [self.objs[3]]

    def test_shallow_obj_lookup(self):
        assert list(JSONModel.objects.filter(attrs__a="b")) == [
            self.objs[1],
            self.objs[2],
        ]

    def test_shallow_obj_lookup_number_key(self):
        assert list(JSONModel.objects.filter(**{'attrs__"9001"': 9002})) == [
            self.objs[1]
        ]

    def test_deep_lookup_objs(self):
        assert list(JSONModel.objects.filter(attrs__k__l="m")) == [self.objs[2]]

    def test_shallow_lookup_obj_target(self):
        assert list(JSONModel.objects.filter(attrs__k={"l": "m"})) == [self.objs[2]]

    def test_deep_lookup_array(self):
        assert list(JSONModel.objects.filter(attrs__1__0=2)) == [self.objs[3]]

    def test_deep_lookup_mixed(self):
        assert list(JSONModel.objects.filter(attrs__d__1__f="g")) == [self.objs[2]]

    def test_deep_lookup_gt(self):
        assert list(JSONModel.objects.filter(attrs__c__gt=1)) == []

    def test_deep_lookup_lt(self):
        assert list(JSONModel.objects.filter(attrs__c__lt=5)) == [
            self.objs[1],
            self.objs[2],
        ]

    def test_usage_in_subquery(self):
        assert list(
            JSONModel.objects.filter(id__in=JSONModel.objects.filter(attrs__c=1))
        ) == [self.objs[1], self.objs[2]]

    def test_in_lookup(self):
        objs = list(JSONModel.objects.filter(attrs__a__in=["x", "b", "y"]))
        assert objs == [self.objs[1], self.objs[2]]

        objs = list(
            JSONModel.objects.filter(
                attrs__d__1__in=[{"x": "g"}, {"f": "g"}, {"f": "x"}]
            )
        )
        assert objs == [self.objs[2]]

        objs = list(JSONModel.objects.filter(attrs__k__in=[True, {"l": "m"}, "x"]))
        assert objs == [self.objs[2], self.objs[4]]

        objs = list(JSONModel.objects.filter(attrs__k__in=[{"l": "m"}, "x"]))
        assert objs == [self.objs[2]]

        objs = list(JSONModel.objects.filter(attrs__k__in=[True, "x"]))
        assert objs == [self.objs[4]]

        with pytest.raises(ValueError) as excinfo:
            JSONModel.objects.filter(attrs__a__in=[F("name"), "n"])
        assert (
            str(excinfo.value)
            == "JSONField's 'in' lookup only works with literal values"
        )

        with pytest.raises(ValueError) as excinfo:
            subquery = JSONModel.objects.filter(name="b").values_list("attrs")
            JSONModel.objects.filter(attrs__in=subquery)
        assert (
            str(excinfo.value)
            == "JSONField's 'in' lookup only works with literal values"
        )


class TestCheck(JSONFieldTestCase):

    if django.VERSION >= (2, 2):
        databases = ["default", "other"]
    else:
        multi_db = True

    def test_mutable_default_list(self):
        class InvalidJSONModel1(TemporaryModel):
            field = JSONField(default=[])

        errors = InvalidJSONModel1.check(actually_check=True)
        assert len(errors) == 1
        assert errors[0].id == "django_mysql.E017"
        assert "Do not use mutable defaults for JSONField" in errors[0].msg

    def test_mutable_default_dict(self):
        class InvalidJSONModel2(TemporaryModel):
            field = JSONField(default={})

        errors = InvalidJSONModel2.check(actually_check=True)
        assert len(errors) == 1
        assert errors[0].id == "django_mysql.E017"
        assert "Do not use mutable defaults for JSONField" in errors[0].msg

    @mock.patch("django_mysql.models.fields.json.connection_is_mariadb")
    def test_db_not_mysql(self, is_mariadb):
        is_mariadb.return_value = True

        class InvalidJSONModel3(TemporaryModel):
            field = JSONField()

        errors = InvalidJSONModel3.check(actually_check=True)
        assert len(errors) == 1
        assert errors[0].id == "django_mysql.E016"
        assert "MySQL 5.7+ is required" in errors[0].msg

    @mock_mysql_version(default=(5, 5, 3), other=(5, 5, 3))
    def test_mysql_old_version(self):
        class InvalidJSONModel4(TemporaryModel):
            field = JSONField()

        errors = InvalidJSONModel4.check(actually_check=True)
        assert len(errors) == 1
        assert errors[0].id == "django_mysql.E016"
        assert "MySQL 5.7+ is required" in errors[0].msg

    @mock_mysql_version(default=(5, 5, 3), other=(5, 7, 1))
    def test_mysql_one_conn_old_version(self):
        class InvalidJSONModel5(TemporaryModel):
            field = JSONField()

        errors = InvalidJSONModel5.check(actually_check=True)
        assert len(errors) == 0

    def test_bad_custom_encoder(self):
        class InvalidJSONModel6(TemporaryModel):
            field = JSONField(encoder=json.JSONEncoder())

        errors = InvalidJSONModel6.check(actually_check=True)
        assert len(errors) == 1
        assert errors[0].id == "django_mysql.E018"
        assert errors[0].msg.startswith("Custom JSON encoder should have")

    def test_bad_custom_decoder(self):
        class InvalidJSONModel7(TemporaryModel):
            field = JSONField(decoder=json.JSONDecoder())

        errors = InvalidJSONModel7.check(actually_check=True)
        assert len(errors) == 1
        assert errors[0].id == "django_mysql.E019"
        assert errors[0].msg.startswith("Custom JSON decoder should have")


class TestFormField(JSONFieldTestCase):
    def test_model_field_formfield(self):
        model_field = JSONField()
        form_field = model_field.formfield()
        assert isinstance(form_field, forms.JSONField)


class JSONFieldSubclass(JSONField):
    pass


class TestDeconstruct(JSONFieldTestCase):
    def test_deconstruct(self):
        field = JSONField()
        name, path, args, kwargs = field.deconstruct()

        new = JSONField(*args, **kwargs)

        assert path == "django_mysql.models.JSONField"
        assert new.default == field.default

    def test_deconstruct_subclass(self):
        field = JSONFieldSubclass()
        name, path, args, kwargs = field.deconstruct()
        assert path == "tests.testapp.test_jsonfield.JSONFieldSubclass"


class TestSerialization(JSONFieldTestCase):
    test_data = """[
        {
            "fields": {
                "attrs": {"a": "b", "c": null},
                "name": ""
            },
            "model": "testapp.jsonmodel",
            "pk": null
        }
    ]"""

    def test_dumping(self):
        instance = JSONModel(attrs={"a": "b", "c": None})
        data = serializers.serialize("json", [instance])
        assert json.loads(data) == json.loads(self.test_data)

    def test_loading(self):
        instances = list(serializers.deserialize("json", self.test_data))
        instance = instances[0].object
        assert instance.attrs == {"a": "b", "c": None}



from django.db import migrations

from django_mysql.models import SizedTextField


class Migration(migrations.Migration):

    dependencies = [("testapp", "0001_initial")]

    operations = [
        migrations.AlterField(
            model_name="sizedtextaltermodel",
            name="field",
            field=SizedTextField(size_class=1),
        )
    ]



import pickle
import re
from unittest import mock, skipUnless

import pytest
from django.contrib.contenttypes.models import ContentType
from django.db import DEFAULT_DB_ALIAS, connections
from django.db.models.query import QuerySet
from django.template import Context, Template
from django.test import TestCase
from django.test.utils import captured_stdout, override_settings

from django_mysql.models import ApproximateInt, SmartIterator, add_QuerySetMixin
from django_mysql.utils import have_program, index_name
from tests.testapp.models import (
    Author,
    AuthorExtra,
    AuthorMultiIndex,
    Book,
    NameAuthor,
    NameAuthorExtra,
    VanillaAuthor,
)
from tests.testapp.utils import CaptureLastQuery, skip_if_mysql_8_plus, used_indexes


class MixinQuerysetTests(TestCase):
    def test_ContentType(self):
        qs = ContentType.objects.all()
        assert not hasattr(qs, "approx_count")
        qs2 = add_QuerySetMixin(qs)

        assert hasattr(qs2, "approx_count")
        assert qs.__class__ != qs2.__class__
        assert qs.__class__ in qs2.__class__.__mro__


class ApproximateCountTests(TestCase):
    def setUp(self):
        super().setUp()
        Author.objects.bulk_create([Author() for i in range(10)])

    def test_activation_deactivation(self):
        qs = Author.objects.all()
        assert not qs._count_tries_approx

        qs2 = qs.count_tries_approx(min_size=2)
        assert qs != qs2
        assert qs2._count_tries_approx
        count = qs2.count()
        assert isinstance(count, ApproximateInt)

        qs3 = qs2.count_tries_approx(False)
        assert qs2 != qs3
        assert not qs3._count_tries_approx

    def test_activation_but_fallback(self):
        qs = Author.objects.exclude(name="TEST").count_tries_approx()
        count = qs.count()
        assert count == 10
        assert not isinstance(count, ApproximateInt)

    def test_activation_but_fallback_due_to_min_size(self):
        qs = Author.objects.count_tries_approx()
        count = qs.count()
        assert count == 10
        assert not isinstance(count, ApproximateInt)

    def test_output_in_templates(self):
        approx_count = Author.objects.approx_count(min_size=1)
        text = Template("{{ var }}").render(Context({"var": approx_count}))
        assert text.startswith("Approximately ")

        approx_count2 = Author.objects.approx_count(min_size=1, return_approx_int=False)
        text = Template("{{ var }}").render(Context({"var": approx_count2}))
        assert not text.startswith("Approximately ")

    def test_fallback_with_filters(self):
        filtered = Author.objects.filter(name="")
        assert filtered.approx_count(fall_back=True) == 10
        with pytest.raises(ValueError):
            filtered.approx_count(fall_back=False)

    def test_fallback_with_slice(self):
        assert Author.objects.all()[:100].approx_count() == 10
        with pytest.raises(ValueError):
            Author.objects.all()[:100].approx_count(fall_back=False)

    def test_fallback_with_distinct(self):
        assert Author.objects.distinct().approx_count() == 10
        with pytest.raises(ValueError):
            Author.objects.distinct().approx_count(fall_back=False)

    def test_fallback_with_arbitrary_extra(self):
        assert Author.objects.extra(where=["1=1"]).approx_count() == 10
        with pytest.raises(ValueError):
            Author.objects.extra(where=["1=1"]).approx_count(fall_back=False)

    def test_approx_count_with_label(self):
        # It should be possible to approx count a query set with a query hint
        # as none of them affect the result
        assert Author.objects.label("bla").approx_count(fall_back=False) == 10

    def test_approx_count_with_straight_join(self):
        # It should be possible to approx count a query set with a query hint
        # as none of them affect the result
        assert Author.objects.straight_join().approx_count(fall_back=False) == 10


class QueryHintTests(TestCase):
    def test_label(self):
        with CaptureLastQuery() as cap:
            list(Author.objects.label("QueryHintTests.test_label").all())
        assert cap.query.startswith("SELECT /*QueryHintTests.test_label*/ ")

    def test_label_twice(self):
        with CaptureLastQuery() as cap:
            list(Author.objects.label("QueryHintTests").label("test_label_twice").all())
        assert cap.query.startswith("SELECT /*QueryHintTests*/ /*test_label_twice*/ ")

    def test_label_star(self):
        with CaptureLastQuery() as cap:
            list(Author.objects.label("I'ma*").label("*").all())
        assert cap.query.startswith("SELECT /*I'ma**/ /***/ ")

    def test_label_update(self):
        Author.objects.create(name="UPDATEME")
        with CaptureLastQuery() as cap:
            Author.objects.label("QueryHintTests").update(name="UPDATED")
        assert cap.query.startswith("UPDATE /*QueryHintTests*/ ")

    def test_label_bad(self):
        with self.assertRaises(ValueError):
            Author.objects.label("badlabel*/")

    def test_label_and_straight_join(self):
        with CaptureLastQuery() as cap:
            list(
                Author.objects.label("QueryHintTests.test_label_and")
                .straight_join()
                .all()
            )
        assert cap.query.startswith(
            "SELECT /*QueryHintTests.test_label_and*/ STRAIGHT_JOIN "
        )

    def test_straight_join(self):
        with CaptureLastQuery() as cap:
            list(Author.objects.filter(books__title__startswith="A").straight_join())

        assert cap.query.startswith("SELECT STRAIGHT_JOIN ")

    def test_straight_join_with_distinct(self):
        with CaptureLastQuery() as cap:
            list(
                Author.objects.filter(tutor=None)
                .distinct()
                .values("books__title")
                .straight_join()
            )

        assert cap.query.startswith("SELECT DISTINCT STRAIGHT_JOIN ")

    @override_settings(DJANGO_MYSQL_REWRITE_QUERIES=False)
    def test_disabled_setting_errors_on_usage(self):
        with pytest.raises(RuntimeError) as excinfo:
            Author.objects.all().straight_join()

        assert "DJANGO_MYSQL_REWRITE_QUERIES" in str(excinfo.value)

    @skip_if_mysql_8_plus()
    def test_sql_cache(self):
        with CaptureLastQuery() as cap:
            list(Author.objects.sql_cache().all())
        assert cap.query.startswith("SELECT SQL_CACHE ")

    @skip_if_mysql_8_plus()
    def test_sql_no_cache(self):
        with CaptureLastQuery() as cap:
            list(Author.objects.sql_no_cache().all())
        assert cap.query.startswith("SELECT SQL_NO_CACHE ")

    def test_sql_small_result(self):
        with CaptureLastQuery() as cap:
            list(Author.objects.sql_small_result().all())
        assert cap.query.startswith("SELECT SQL_SMALL_RESULT ")

    def test_sql_big_result(self):
        with CaptureLastQuery() as cap:
            list(Author.objects.sql_big_result().all())
        assert cap.query.startswith("SELECT SQL_BIG_RESULT ")

    def test_sql_buffer_result(self):
        with CaptureLastQuery() as cap:
            list(Author.objects.sql_buffer_result().all())
        assert cap.query.startswith("SELECT SQL_BUFFER_RESULT ")

    def test_adding_many(self):
        with CaptureLastQuery() as cap:
            list(Author.objects.straight_join().sql_big_result().sql_buffer_result())
        assert cap.query.startswith(
            "SELECT STRAIGHT_JOIN SQL_BIG_RESULT SQL_BUFFER_RESULT "
        )

    def test_complex_query_1(self):
        with CaptureLastQuery() as cap:
            list(
                Author.objects.distinct()
                .straight_join()
                .filter(books__title__startswith="A")
                .exclude(books__id__lte=1)
                .prefetch_related("tutees")
                .filter(bio__gt="")
                .exclude(bio__startswith="Once upon")
            )
        assert cap.query.startswith("SELECT DISTINCT STRAIGHT_JOIN ")

    def test_complex_query_2(self):
        subq = Book.objects.straight_join().filter(title__startswith="A")
        with CaptureLastQuery() as cap:
            list(Author.objects.straight_join().filter(books__in=subq))
        assert cap.query.startswith("SELECT STRAIGHT_JOIN ")

    def test_use_index(self):
        name_idx = index_name(Author, "name")
        with CaptureLastQuery() as cap:
            list(Author.objects.filter(name__gt="").use_index(name_idx))
        assert ("USE INDEX (`" + name_idx + "`)") in cap.query
        used = used_indexes(cap.query)
        assert len(used) == 0 or name_idx in used

    def test_use_index_primary(self):
        with CaptureLastQuery() as cap:
            list(Author.objects.use_index("PRIMARY"))
        assert ("USE INDEX (`PRIMARY`)") in cap.query
        used = used_indexes(cap.query)
        assert len(used) == 0 or "PRIMARY" in used

    def test_force_index(self):
        name_idx = index_name(Author, "name")
        with CaptureLastQuery() as cap:
            list(Author.objects.filter(name__gt="").force_index(name_idx))
        assert ("FORCE INDEX (`" + name_idx + "`)") in cap.query
        assert name_idx in used_indexes(cap.query)

    def test_force_index_primary(self):
        with CaptureLastQuery() as cap:
            list(Author.objects.force_index("PRIMARY"))
        assert ("FORCE INDEX (`PRIMARY`)") in cap.query
        used = used_indexes(cap.query)
        assert len(used) == 0 or "PRIMARY" in used

    def test_ignore_index(self):
        name_idx = index_name(Author, "name")
        with CaptureLastQuery() as cap:
            list(Author.objects.filter(name__gt="").ignore_index(name_idx))
        assert ("IGNORE INDEX (`" + name_idx + "`)") in cap.query
        assert name_idx not in used_indexes(cap.query)

    def test_ignore_index_multiple(self):
        name_idx = index_name(AuthorMultiIndex, "name")
        name_country_idx = index_name(AuthorMultiIndex, "name", "country")
        with CaptureLastQuery() as cap:
            list(
                AuthorMultiIndex.objects.filter(name__gt="").ignore_index(
                    name_idx, name_country_idx
                )
            )
        assert (
            "IGNORE INDEX (`" + name_idx + "`,`" + name_country_idx + "`)"
        ) in cap.query
        used = used_indexes(cap.query)
        assert name_idx not in used
        assert name_country_idx not in used

    def test_ignore_index_primary(self):
        with CaptureLastQuery() as cap:
            list(Author.objects.filter(name__gt="").ignore_index("PRIMARY"))
        assert "IGNORE INDEX (`PRIMARY`)" in cap.query
        assert "PRIMARY" not in used_indexes(cap.query)

    def test_force_index_at_least_one(self):
        with pytest.raises(ValueError) as excinfo:
            Author.objects.force_index()
        assert str(excinfo.value) == "force_index requires at least one index name"

    def test_force_index_invalid_for(self):
        with pytest.raises(ValueError) as excinfo:
            Author.objects.force_index("a", for_="INVALID")
        assert "for_ must be one of" in str(excinfo.value)

    def test_force_index_invalid_kwarg(self):
        with pytest.raises(ValueError) as excinfo:
            Author.objects.force_index("a", nonexistent=True)
        assert (
            "force_index accepts only 'for_' and 'table_name' as keyword arguments"
            in str(excinfo.value)
        )

    def test_index_hint_force_order_by(self):
        name_idx = index_name(Author, "name")
        with CaptureLastQuery() as cap:
            list(Author.objects.force_index(name_idx, for_="ORDER BY").order_by("name"))

        assert ("FORCE INDEX FOR ORDER BY (`" + name_idx + "`)") in cap.query
        assert name_idx in used_indexes(cap.query)

    def test_use_index_none(self):
        with CaptureLastQuery() as cap:
            list(Author.objects.values_list("name").distinct().use_index())
        assert "USE INDEX () " in cap.query
        assert used_indexes(cap.query) == set()

    def test_use_index_table_name(self):
        extra_table = "testapp_authorextra"
        with CaptureLastQuery() as cap:
            list(
                Author.objects.select_related("authorextra").use_index(
                    "PRIMARY", table_name=extra_table
                )
            )
        assert ("`" + extra_table + "` USE INDEX (`PRIMARY`) ") in cap.query

    def test_force_index_table_name_doesnt_exist_ignored(self):
        with CaptureLastQuery() as cap:
            list(
                Author.objects.select_related("authorextra").force_index(
                    "PRIMARY", table_name="nonexistent"
                )
            )
        assert " FORCE INDEX " not in cap.query


class QueryHintNewConnectionTests(TestCase):
    @classmethod
    def setUpClass(cls):
        # Force a new connection to be created, in order to check that the
        # rewrite hook is installed on it.
        del connections[DEFAULT_DB_ALIAS]
        super().setUpClass()

    def test_label_forced_new_connection(self):
        with CaptureLastQuery() as cap:
            list(Author.objects.label("QueryHintTests.test_label").all())
        assert cap.query.startswith("SELECT /*QueryHintTests.test_label*/ ")


class FoundRowsTests(TestCase):
    def setUp(self):
        super().setUp()
        Author.objects.bulk_create([Author() for i in range(10)])

    def test_found_rows_requires_sql_calc_found_rows(self):
        authors = Author.objects.all()
        with pytest.raises(ValueError) as excinfo:
            authors.found_rows
        assert "can only be used if" in str(excinfo.value)

    def test_found_rows_requires_iteration(self):
        authors = Author.objects.sql_calc_found_rows()[:5]
        with self.assertRaises(RuntimeError):
            authors.found_rows

    def test_it_doesnt_work_with_iterator(self):
        authors = Author.objects.sql_calc_found_rows()[:5]
        match_re = r"doesn't work with iterator\(\)"
        with pytest.raises(ValueError, match=match_re):
            authors.iterator()

    def test_iterator_without_it_works(self):
        authors = Author.objects.all()[:5]
        authors.iterator()

    def test_it_working(self):
        with self.assertNumQueries(2):
            authors = Author.objects.sql_calc_found_rows()[:5]
            list(authors)
            assert authors.found_rows == 10

    def test_no_repeats(self):
        with self.assertNumQueries(2):
            authors = Author.objects.sql_calc_found_rows().sql_calc_found_rows()[:5]
            list(authors)
            assert authors.found_rows == 10

    def test_it_working_prefetching(self):
        with self.assertNumQueries(3):
            authors = Author.objects.sql_calc_found_rows().prefetch_related("tutor")[:5]
            list(authors)
            assert authors.found_rows == 10

    def test_pickleability(self):
        authors = Author.objects.sql_calc_found_rows()
        pickle.dumps(authors)

    def test_cloning_doesnt_copy_found_rows_number(self):
        authors = Author.objects.sql_calc_found_rows()
        list(authors)
        assert authors.found_rows == 10

        authors2 = authors.filter(id__gte=1)
        with pytest.raises(RuntimeError):
            authors2.found_rows

        with self.assertNumQueries(2):
            list(authors2)
            assert authors2.found_rows == 10


class SmartIteratorTests(TestCase):
    def setUp(self):
        super().setUp()
        Author.objects.bulk_create([Author(id=i + 1) for i in range(10)])

    def test_bad_querysets(self):
        with pytest.raises(ValueError) as excinfo:
            Author.objects.all().order_by("name").iter_smart_chunks()
        assert "ordering" in str(excinfo.value)

        with pytest.raises(ValueError) as excinfo:
            Author.objects.all()[:5].iter_smart_chunks()
        assert "sliced QuerySet" in str(excinfo.value)

        with pytest.raises(ValueError) as excinfo:
            NameAuthor.objects.all().iter_smart_chunks()
        assert "non-integer primary key" in str(excinfo.value)

    def test_chunks(self):
        seen = []
        for authors in Author.objects.iter_smart_chunks():
            seen.extend(author.id for author in authors)

        all_ids = list(Author.objects.order_by("id").values_list("id", flat=True))
        assert seen == all_ids

    def test_objects(self):
        seen = [author.id for author in Author.objects.iter_smart()]
        all_ids = list(Author.objects.order_by("id").values_list("id", flat=True))
        assert seen == all_ids

    def test_objects_non_atomic(self):
        seen = [a.id for a in Author.objects.iter_smart(atomically=False)]
        all_ids = list(Author.objects.order_by("id").values_list("id", flat=True))
        assert seen == all_ids

    def test_objects_reverse(self):
        seen = [a.id for a in Author.objects.reverse().iter_smart()]
        all_ids = list(Author.objects.order_by("-id").values_list("id", flat=True))
        assert seen == all_ids

    def test_objects_pk_range_all(self):
        seen = [a.id for a in Author.objects.iter_smart(pk_range="all")]
        all_ids = list(Author.objects.order_by("id").values_list("id", flat=True))
        assert seen == all_ids

    def test_objects_pk_range_tuple(self):
        min_id = Author.objects.earliest("id").id
        max_id = Author.objects.order_by("id")[5].id

        seen = [a.id for a in Author.objects.iter_smart(pk_range=(min_id, max_id))]
        cut_ids = list(
            Author.objects.order_by("id")
            .filter(id__gte=min_id, id__lte=max_id)
            .values_list("id", flat=True)
        )
        assert seen == cut_ids

    def test_objects_pk_range_tuple_0_0(self):
        seen = [a.id for a in Author.objects.iter_smart(pk_range=(0, 0))]
        assert seen == []

    def test_objects_pk_range_reversed(self):
        min_id = Author.objects.earliest("id").id
        max_id = Author.objects.latest("id").id

        seen = [
            author.id for author in Author.objects.iter_smart(pk_range=(max_id, min_id))
        ]
        all_ids = list(Author.objects.order_by("-id").values_list("id", flat=True))
        assert seen == all_ids

    def test_objects_pk_range_bad(self):
        with pytest.raises(ValueError) as excinfo:
            list(Author.objects.iter_smart(pk_range="My Bad Value"))
        assert "Unrecognized value for pk_range" in str(excinfo.value)

    def test_pk_range_race_condition(self):
        getitem = QuerySet.__getitem__

        def fail_second_slice(*args, **kwargs):
            # Simulate race condition by deleting all objects between first
            # call (min_qs[0]) and second call (max_qs[0]) to
            # QuerySet.__getitem__
            fail_second_slice.calls += 1
            if fail_second_slice.calls == 2:
                Author.objects.all().delete()
            return getitem(*args, **kwargs)

        fail_second_slice.calls = 0

        path = "django.db.models.query.QuerySet.__getitem__"

        with mock.patch(path, fail_second_slice):
            seen = [author.id for author in Author.objects.iter_smart()]
        assert seen == []

    def test_objects_chunk_size(self):
        smart = iter(Author.objects.iter_smart_chunks(chunk_size=3))
        chunk = next(smart)
        assert len(list(chunk)) == 3

    def test_objects_chunk_size_1(self):
        smart = iter(Author.objects.iter_smart_chunks(chunk_size=1))
        chunk = next(smart)
        assert len(list(chunk)) == 1

    def test_objects_max_size(self):
        seen = []
        for chunk in Author.objects.iter_smart_chunks(chunk_max=3):
            ids = [author.id for author in chunk]
            assert len(ids) <= 3
            seen.extend(ids)
        all_ids = list(Author.objects.order_by("id").values_list("id", flat=True))
        assert seen == all_ids

    def test_objects_max_size_1(self):
        seen = []
        for chunk in Author.objects.iter_smart_chunks(chunk_max=1):
            ids = [author.id for author in chunk]
            assert len(ids) == 1
            seen.extend(ids)
        all_ids = list(Author.objects.order_by("id").values_list("id", flat=True))
        assert seen == all_ids

    def test_objects_max_size_bounds_chunk_size(self):
        smart = iter(Author.objects.iter_smart_chunks(chunk_max=5, chunk_size=1000))

        seen = []
        for chunk in smart:
            ids = [author.id for author in chunk]
            assert len(ids) <= 5
            seen.extend(ids)
        all_ids = list(Author.objects.order_by("id").values_list("id", flat=True))
        assert seen == all_ids

    def test_objects_min_size_bounds_chunk_size(self):
        smart = iter(Author.objects.iter_smart_chunks(chunk_min=5, chunk_size=1))
        seen = []
        for chunk in smart:
            ids = [author.id for author in chunk]
            assert len(ids) >= 5
            seen.extend(ids)
        all_ids = list(Author.objects.order_by("id").values_list("id", flat=True))
        assert seen == all_ids

    def test_min_size_greater_than_max(self):
        with pytest.raises(AssertionError):
            Author.objects.iter_smart_chunks(chunk_min=2, chunk_max=1)

    def test_no_matching_objects(self):
        seen = [author.id for author in Author.objects.filter(name="Waaa").iter_smart()]
        assert seen == []

    def test_no_objects(self):
        Author.objects.all().delete()
        seen = [author.id for author in Author.objects.iter_smart()]
        assert seen == []

    def test_pk_hole(self):
        first = Author.objects.earliest("id")
        last = Author.objects.latest("id")
        Author.objects.filter(id__gt=first.id, id__lt=last.id).delete()
        seen = [author.id for author in Author.objects.iter_smart()]
        assert seen == [first.id, last.id]

    def test_iter_smart_pk_range(self):
        seen = []
        for start_pk, end_pk in Author.objects.iter_smart_pk_ranges():
            seen.extend(
                Author.objects.filter(id__gte=start_pk, id__lt=end_pk).values_list(
                    "id", flat=True
                )
            )
        all_ids = list(Author.objects.order_by("id").values_list("id", flat=True))
        assert seen == all_ids

    def test_iter_smart_pk_range_with_raw(self):
        seen = []
        for start_pk, end_pk in Author.objects.iter_smart_pk_ranges():
            authors = Author.objects.raw(
                """
                SELECT id FROM {}
                WHERE id >= %s AND id < %s
            """.format(
                    Author._meta.db_table
                ),
                (start_pk, end_pk),
            )
            seen.extend(author.id for author in authors)

        all_ids = list(Author.objects.order_by("id").values_list("id", flat=True))
        assert seen == all_ids

    def test_iter_smart_fk_primary_key(self):
        author, author2 = Author.objects.all()[:2]
        AuthorExtra.objects.create(author=author, legs=2)
        AuthorExtra.objects.create(author=author2, legs=1)

        seen_author_ids = []
        for extra in AuthorExtra.objects.iter_smart():
            seen_author_ids.append(extra.author_id)
        assert seen_author_ids == [author.id, author2.id]

    def test_iter_smart_fk_string_primary_key_fails(self):
        with pytest.raises(ValueError) as excinfo:
            NameAuthorExtra.objects.iter_smart()
        assert "non-integer primary key" in str(excinfo.value)

    def test_reporting(self):
        with captured_stdout() as output:
            qs = Author.objects.all()
            for authors in qs.iter_smart_chunks(report_progress=True):
                list(authors)  # fetch them

        lines = output.getvalue().split("\n")

        reports = lines[0].split("\r")
        for report in reports:
            assert re.match(
                r"^Author SmartChunkedIterator processed \d+/10 objects "
                r"\(\d+\.\d+%\) in \d+ chunks"
                r"(; highest pk so far \d+(, [\dhms]+ remaining)?)?$",
                report,
            )

        assert re.match(r"Finished! Iterated over \d+ objects? in [\dhms]+.", lines[1])

    def test_reporting_reverse(self):
        Author.objects.create(id=10000)

        with captured_stdout() as output:
            qs = Author.objects.reverse()
            for authors in qs.iter_smart_chunks(report_progress=True, chunk_min=100):
                list(authors)  # fetch them

        lines = output.getvalue().split("\n")

        reports = lines[0].split("\r")
        assert len(reports) > 0
        for report in reports:
            assert re.match(
                r"^Author SmartChunkedIterator processed \d+/11 objects "
                r"\(\d+\.\d+%\) in \d+ chunks"
                r"(; lowest pk so far \d+(, [\dhms]+ remaining)?)?( )*$",
                report,
            )

        assert re.match(r"Finished! Iterated over \d+ objects? in [\dhms]+.", lines[1])

    def test_reporting_with_total(self):
        with captured_stdout() as output:
            qs = Author.objects.all()
            for authors in qs.iter_smart_chunks(report_progress=True, total=4):
                list(authors)  # fetch them

        lines = output.getvalue().split("\n")

        reports = lines[0].split("\r")
        for report in reports:
            assert re.match(
                r"^Author SmartChunkedIterator processed \d+/4 objects "
                r"\(\d+\.\d+%\) in \d+ chunks"
                r"(; highest pk so far \d+(, [\dhms]+ remaining)?)?$",
                report,
            )

        assert re.match(r"Finished! Iterated over \d+ objects? in [\dhms]+.", lines[1])

    def test_reporting_on_uncounted_qs(self):
        Author.objects.create(name="pants")

        with captured_stdout() as output:
            qs = Author.objects.filter(name="pants")
            for authors in qs.iter_smart_chunks(report_progress=True):
                authors.delete()

        lines = output.getvalue().split("\n")

        reports = lines[0].split("\r")
        for report in reports:
            assert re.match(
                # We should have ??? since the deletion means the objects
                # aren't fetched into python
                r"Author SmartChunkedIterator processed (0|\?\?\?)/1 objects "
                r"\(\d+\.\d+%\) in \d+ chunks(; highest pk so far \d+)?",
                report,
            )

        assert re.match(
            r"Finished! Iterated over \?\?\? objects? in [\dhms]+.", lines[1]
        )

    def test_filter_and_delete(self):
        VanillaAuthor.objects.create(name="Alpha")
        VanillaAuthor.objects.create(name="pants")
        VanillaAuthor.objects.create(name="Beta")
        VanillaAuthor.objects.create(name="pants")

        bad_authors = VanillaAuthor.objects.filter(name="pants")

        assert bad_authors.count() == 2

        with captured_stdout():
            for author in SmartIterator(bad_authors, report_progress=True):
                author.delete()

        assert bad_authors.count() == 0


@skipUnless(have_program("pt-visual-explain"), "pt-visual-explain must be installed")
class VisualExplainTests(TestCase):
    def test_basic(self):
        with captured_stdout() as capture:
            Author.objects.all().pt_visual_explain()
        output = capture.getvalue()
        # Can't be too strict about the output since different database and pt-
        # visual-explain versions give different output
        assert "testapp_author" in output
        assert "rows" in output
        assert "Table" in output

    def test_basic_no_display(self):
        output = Author.objects.all().pt_visual_explain(display=False)
        assert "testapp_author" in output
        assert "rows" in output
        assert "Table" in output

    def test_subquery(self):
        subq = Author.objects.all().values_list("id", flat=True)
        output = Author.objects.filter(id__in=subq).pt_visual_explain(display=False)
        assert "possible_keys" in output
        assert "testapp_author" in output
        assert "rows" in output
        assert "Table" in output



from django.db import migrations

from django_mysql.models import SizedBinaryField


class Migration(migrations.Migration):

    dependencies = [("testapp", "0001_initial")]

    operations = [
        migrations.AlterField(
            model_name="sizedbinaryaltermodel",
            name="field",
            field=SizedBinaryField(size_class=2),
        )
    ]



import django
from django.core.management import call_command
from django.test import TestCase, TransactionTestCase

from django_mysql.checks import check_variables
from django_mysql.test.utils import override_mysql_variables


class CallCheckTest(TestCase):

    if django.VERSION >= (2, 2):
        databases = ["default", "other"]
    else:
        multi_db = True

    def test_check(self):
        call_command("check")


class VariablesTests(TransactionTestCase):

    if django.VERSION >= (2, 2):
        databases = ["default", "other"]
    else:
        multi_db = True

    def test_passes(self):
        assert check_variables([]) == []

    @override_mysql_variables(sql_mode="")
    def test_fails_if_no_strict(self):
        errors = check_variables([])
        assert len(errors) == 1
        assert errors[0].id == "django_mysql.W001"
        assert "MySQL Strict Mode" in errors[0].msg

    @override_mysql_variables(innodb_strict_mode=0)
    def test_fails_if_no_innodb_strict(self):
        errors = check_variables([])
        assert len(errors) == 1
        assert errors[0].id == "django_mysql.W002"
        assert "InnoDB Strict Mode" in errors[0].msg

    @override_mysql_variables(character_set_connection="utf8")
    def test_fails_if_not_utf8mb4(self):
        errors = check_variables([])
        assert len(errors) == 1
        assert errors[0].id == "django_mysql.W003"
        assert "utf8mb4" in errors[0].msg



import django
import pytest
from django.db import connections
from django.test import TestCase

from django_mysql.test.utils import override_mysql_variables


class OverrideVarsMethodTest(TestCase):
    @override_mysql_variables(TIMESTAMP=123)
    def test_method_decorator(self):
        self.check_timestamp(123)

    def check_timestamp(self, expected, using="default"):
        with connections[using].cursor() as cursor:
            cursor.execute("SELECT @@TIMESTAMP")
            mode = cursor.fetchone()[0]

        assert mode == expected


@override_mysql_variables(TIMESTAMP=123)
class OverrideVarsClassTest(OverrideVarsMethodTest):

    if django.VERSION >= (2, 2):
        databases = ["default", "other"]
    else:
        multi_db = True

    def test_class_decorator(self):
        self.check_timestamp(123)

    @override_mysql_variables(using="other", TIMESTAMP=456)
    def test_other_connection(self):
        self.check_timestamp(123)
        self.check_timestamp(456, using="other")

    def test_it_fails_on_non_test_classes(self):
        with pytest.raises(Exception):

            @override_mysql_variables(TIMESTAMP=123)
            class MyClass:
                pass
